from __future__ import annotations

import inspect
import logging
import os
import warnings
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING
from uuid import uuid4

import numpy as np
import rich
import torch
from anndata import AnnData
from mudata import MuData

from scvi import REGISTRY_KEYS, settings
from scvi.data import AnnDataManager, fields
from scvi.data._compat import registry_from_setup_dict
from scvi.data._constants import (
    _ADATA_MINIFY_TYPE_UNS_KEY,
    _MODEL_NAME_KEY,
    _SCVI_UUID_KEY,
    _SETUP_ARGS_KEY,
    _SETUP_METHOD_NAME,
    ADATA_MINIFY_TYPE,
)
from scvi.data._utils import _assign_adata_uuid, _check_if_view, _get_adata_minify_type
from scvi.dataloaders import AnnDataLoader
from scvi.model._utils import parse_device_args
from scvi.model.base._constants import SAVE_KEYS
from scvi.model.base._save_load import (
    _initialize_model,
    _load_legacy_saved_files,
    _load_saved_files,
    _validate_var_names,
)
from scvi.model.utils import get_minified_adata_scrna, get_minified_mudata
from scvi.utils import attrdict, setup_anndata_dsp
from scvi.utils._docstrings import devices_dsp

if TYPE_CHECKING:
    from collections.abc import Sequence

    from scvi._types import AnnOrMuData, MinifiedDataType

logger = logging.getLogger(__name__)


_UNTRAINED_WARNING_MESSAGE = (
    "Trying to query inferred values from an untrained model. Please train the model first."
)

_SETUP_INPUTS_EXCLUDED_PARAMS = {"adata", "mdata", "kwargs"}


class BaseModelMetaClass(ABCMeta):
    """Metaclass for :class:`~scvi.model.base.BaseModelClass`.

    Constructs model class-specific mappings for :class:`~scvi.data.AnnDataManager` instances.
    ``cls._setup_adata_manager_store`` maps from AnnData object UUIDs to
    :class:`~scvi.data.AnnDataManager` instances.

    This mapping is populated everytime ``cls.setup_anndata()`` is called.
    ``cls._per_isntance_manager_store`` maps from model instance UUIDs to AnnData UUID:
    :class:`~scvi.data.AnnDataManager` mappings.
    These :class:`~scvi.data.AnnDataManager` instances are tied to a single model instance and
    populated either
    during model initialization or after running ``self._validate_anndata()``.
    """

    @abstractmethod
    def __init__(cls, name, bases, dct):
        cls._setup_adata_manager_store: dict[
            str, type[AnnDataManager]
        ] = {}  # Maps adata id to AnnDataManager instances.
        cls._per_instance_manager_store: dict[
            str, dict[str, type[AnnDataManager]]
        ] = {}  # Maps model instance id to AnnDataManager mappings.
        super().__init__(name, bases, dct)


class BaseModelClass(metaclass=BaseModelMetaClass):
    """Abstract class for scvi-tools models.

    Notes
    -----
    See further usage examples in the following tutorials:

    1. :doc:`/tutorials/notebooks/dev/model_user_guide`
    """

    _OBSERVED_LIB_SIZE_KEY = "observed_lib_size"
    _data_loader_cls = AnnDataLoader

    def __init__(self, adata: AnnOrMuData | None = None):
        # check if the given adata is minified and check if the model being created
        # supports minified-data mode (i.e. inherits from the abstract BaseMinifiedModeModelClass).
        # If not, raise an error to inform the user of the lack of minified-data functionality
        # for this model
        data_is_minified = adata is not None and _get_adata_minify_type(adata) is not None
        if data_is_minified and not issubclass(type(self), BaseMinifiedModeModelClass):
            raise NotImplementedError(
                f"The {type(self).__name__} model currently does not support minified data."
            )
        self.id = str(uuid4())  # Used for cls._manager_store keys.
        if adata is not None:
            self._adata = adata
            self._adata_manager = self._get_most_recent_anndata_manager(adata, required=True)
            self._register_manager_for_instance(self.adata_manager)
            # Suffix registry instance variable with _ to include it when saving the model.
            self.registry_ = self._adata_manager.registry
            self.summary_stats = self._adata_manager.summary_stats

        self._module_init_on_train = adata is None
        self.is_trained_ = False
        self._model_summary_string = ""
        self.train_indices_ = None
        self.test_indices_ = None
        self.validation_indices_ = None
        self.history_ = None

    @property
    def adata(self) -> AnnOrMuData:
        """Data attached to model instance."""
        return self._adata

    @adata.setter
    def adata(self, adata: AnnOrMuData):
        if adata is None:
            raise ValueError("adata cannot be None.")
        self._validate_anndata(adata)
        self._adata = adata
        self._adata_manager = self.get_anndata_manager(adata, required=True)
        self.registry_ = self._adata_manager.registry
        self.summary_stats = self._adata_manager.summary_stats

    @property
    def adata_manager(self) -> AnnDataManager:
        """Manager instance associated with self.adata."""
        return self._adata_manager

    def to_device(self, device: str | int):
        """Move model to device.

        Parameters
        ----------
        device
            Device to move model to. Options: 'cpu' for CPU, integer GPU index (eg. 0),
            or 'cuda:X' where X is the GPU index (eg. 'cuda:0'). See torch.device for more info.

        Examples
        --------
        >>> adata = scvi.data.synthetic_iid()
        >>> model = scvi.model.SCVI(adata)
        >>> model.to_device("cpu")  # moves model to CPU
        >>> model.to_device("cuda:0")  # moves model to GPU 0
        >>> model.to_device(0)  # also moves model to GPU 0
        """
        my_device = torch.device(device)
        self.module.to(my_device)

    @property
    def device(self) -> str:
        """The current device that the module's params are on."""
        return self.module.device

    @staticmethod
    def _get_setup_method_args(**setup_locals) -> dict:
        """Returns a dictionary organizing the arguments used to call ``setup_anndata``.

        Must be called with ``**locals()`` at the start of the ``setup_anndata`` method
        to avoid the inclusion of any extraneous variables.
        """
        cls = setup_locals.pop("cls")
        method_name = None
        if "adata" in setup_locals:
            method_name = "setup_anndata"
        elif "mdata" in setup_locals:
            method_name = "setup_mudata"

        model_name = cls.__name__
        setup_args = {}
        for k, v in setup_locals.items():
            if k not in _SETUP_INPUTS_EXCLUDED_PARAMS:
                setup_args[k] = v
        return {
            _MODEL_NAME_KEY: model_name,
            _SETUP_METHOD_NAME: method_name,
            _SETUP_ARGS_KEY: setup_args,
        }

    @staticmethod
    def _create_modalities_attr_dict(
        modalities: dict[str, str], setup_method_args: dict
    ) -> attrdict:
        """Preprocesses a ``modalities`` dictionary to map modality names.

        Ensures each field key has a respective modality key, defaulting to ``None``.
        Raises a ``UserWarning`` if extraneous modality mappings are detected.

        Parameters
        ----------
        modalities
            Dictionary mapping ``setup_mudata()`` argument name to modality name.
        setup_method_args
            Output of  ``_get_setup_method_args()``.
        """
        setup_args = setup_method_args[_SETUP_ARGS_KEY]
        filtered_modalities = {
            arg_name: modalities.get(arg_name, None) for arg_name in setup_args.keys()
        }
        extra_modalities = set(modalities) - set(filtered_modalities)
        if len(extra_modalities) > 0:
            raise ValueError(f"Extraneous modality mapping(s) detected: {extra_modalities}")
        return attrdict(filtered_modalities)

    @classmethod
    def register_manager(cls, adata_manager: AnnDataManager):
        """Registers an :class:`~scvi.data.AnnDataManager` instance with this model class.

        Stores the :class:`~scvi.data.AnnDataManager` reference in a class-specific manager store.
        Intended for use in the ``setup_anndata()`` class method followed up by retrieval of the
        :class:`~scvi.data.AnnDataManager` via the ``_get_most_recent_anndata_manager()`` method in
        the model init method.

        Notes
        -----
        Subsequent calls to this method with an :class:`~scvi.data.AnnDataManager` instance
        referring to the same underlying AnnData object will overwrite the reference to previous
        :class:`~scvi.data.AnnDataManager`.
        """
        adata_id = adata_manager.adata_uuid
        cls._setup_adata_manager_store[adata_id] = adata_manager

    def _register_manager_for_instance(self, adata_manager: AnnDataManager):
        """Registers an :class:`~scvi.data.AnnDataManager` instance with this model instance.

        Creates a model-instance specific mapping in ``cls._per_instance_manager_store`` for this
        :class:`~scvi.data.AnnDataManager` instance.
        """
        if self.id not in self._per_instance_manager_store:
            self._per_instance_manager_store[self.id] = {}

        adata_id = adata_manager.adata_uuid
        instance_manager_store = self._per_instance_manager_store[self.id]
        instance_manager_store[adata_id] = adata_manager

    def deregister_manager(self, adata: AnnData | None = None):
        """Deregisters the :class:`~scvi.data.AnnDataManager` instance associated with `adata`.

        If `adata` is `None`, deregisters all :class:`~scvi.data.AnnDataManager` instances
        in both the class and instance-specific manager stores, except for the one associated
        with this model instance.
        """
        cls_manager_store = self._setup_adata_manager_store
        instance_manager_store = self._per_instance_manager_store[self.id]

        if adata is None:
            instance_managers_to_clear = list(instance_manager_store.keys())
            cls_managers_to_clear = list(cls_manager_store.keys())
        else:
            adata_manager = self._get_most_recent_anndata_manager(adata, required=True)
            cls_managers_to_clear = [adata_manager.adata_uuid]
            instance_managers_to_clear = [adata_manager.adata_uuid]

        for adata_id in cls_managers_to_clear:
            # don't clear the current manager by default
            is_current_adata = adata is None and adata_id == self.adata_manager.adata_uuid
            if is_current_adata or adata_id not in cls_manager_store:
                continue
            del cls_manager_store[adata_id]

        for adata_id in instance_managers_to_clear:
            # don't clear the current manager by default
            is_current_adata = adata is None and adata_id == self.adata_manager.adata_uuid
            if is_current_adata or adata_id not in instance_manager_store:
                continue
            del instance_manager_store[adata_id]

    @classmethod
    def _get_most_recent_anndata_manager(
        cls, adata: AnnOrMuData, required: bool = False
    ) -> AnnDataManager | None:
        """Retrieves the :class:`~scvi.data.AnnDataManager` for a given AnnData object.

        Checks for the most recent :class:`~scvi.data.AnnDataManager` created for the given AnnData
        object via ``setup_anndata()`` on model initialization. Unlike
        :meth:`scvi.model.base.BaseModelClass.get_anndata_manager`, this method is not model
        instance specific and can be called before a model is fully initialized.

        Parameters
        ----------
        adata
            AnnData object to find manager instance for.
        required
            If True, errors on missing manager. Otherwise, returns None when manager is missing.
        """
        if _SCVI_UUID_KEY not in adata.uns:
            if required:
                raise ValueError(
                    f"Please set up your AnnData with {cls.__name__}.setup_anndata first."
                )
            return None

        adata_id = adata.uns[_SCVI_UUID_KEY]

        if adata_id not in cls._setup_adata_manager_store:
            if required:
                raise ValueError(
                    f"Please set up your AnnData with {cls.__name__}.setup_anndata first. "
                    "It appears the AnnData object has been setup with a different model."
                )
            return None

        adata_manager = cls._setup_adata_manager_store[adata_id]
        if adata_manager.adata is not adata:
            raise ValueError(
                "The provided AnnData object does not match the AnnData object "
                "previously provided for setup. Did you make a copy?"
            )

        return adata_manager

    def get_anndata_manager(
        self, adata: AnnOrMuData, required: bool = False
    ) -> AnnDataManager | None:
        """Retrieves the :class:`~scvi.data.AnnDataManager` for a given AnnData object.

        Requires ``self.id`` has been set. Checks for an :class:`~scvi.data.AnnDataManager`
        specific to this model instance.

        Parameters
        ----------
        adata
            AnnData object to find manager instance for.
        required
            If True, errors on missing manager. Otherwise, returns None when manager is missing.
        """
        cls = self.__class__
        if _SCVI_UUID_KEY not in adata.uns:
            if required:
                raise ValueError(
                    f"Please set up your AnnData with {cls.__name__}.setup_anndata'"
                    "or {cls.__name__}.setup_mudata first."
                )
            return None

        adata_id = adata.uns[_SCVI_UUID_KEY]
        if self.id not in cls._per_instance_manager_store:
            if required:
                raise AssertionError(
                    "Unable to find instance specific manager store. "
                    "The model has likely not been initialized with an AnnData object."
                )
            return None
        elif adata_id not in cls._per_instance_manager_store[self.id]:
            if required:
                raise AssertionError(
                    "Please call ``self._validate_anndata`` on this AnnData or MuData object."
                )
            return None

        adata_manager = cls._per_instance_manager_store[self.id][adata_id]
        if adata_manager.adata is not adata:
            logger.info("AnnData object appears to be a copy. Attempting to transfer setup.")
            _assign_adata_uuid(adata, overwrite=True)
            adata_manager = self.adata_manager.transfer_fields(adata)
            self._register_manager_for_instance(adata_manager)

        return adata_manager

    def get_from_registry(
        self,
        adata: AnnOrMuData,
        registry_key: str,
    ) -> np.ndarray:
        """Returns the object in AnnData associated with the key in the data registry.

        AnnData object should be registered with the model prior to calling this function
        via the ``self._validate_anndata`` method.

        Parameters
        ----------
        registry_key
            key of object to get from data registry.
        adata
            AnnData to pull data from.

        Returns
        -------
        The requested data as a NumPy array.
        """
        adata_manager = self.get_anndata_manager(adata)
        if adata_manager is None:
            raise AssertionError(
                "AnnData not registered with model. Call `self._validate_anndata` "
                "prior to calling this function."
            )
        return adata_manager.get_from_registry(registry_key)

    def _make_data_loader(
        self,
        adata: AnnOrMuData,
        indices: Sequence[int] | None = None,
        batch_size: int | None = None,
        shuffle: bool = False,
        data_loader_class=None,
        **data_loader_kwargs,
    ):
        """Create a AnnDataLoader object for data iteration.

        Parameters
        ----------
        adata
            AnnData object with equivalent structure to initial AnnData.
        indices
            Indices of cells in adata to use. If `None`, all cells are used.
        batch_size
            Minibatch size for data loading into model. Defaults to `scvi.settings.batch_size`.
        shuffle
            Whether observations are shuffled each iteration though
        data_loader_class
            Class to use for data loader
        data_loader_kwargs
            Kwargs to the class-specific data loader class
        """
        adata_manager = self.get_anndata_manager(adata)
        if adata_manager is None:
            raise AssertionError(
                "AnnDataManager not found. Call `self._validate_anndata` prior to calling this "
                "function."
            )

        adata = adata_manager.adata

        if batch_size is None:
            batch_size = settings.batch_size
        if indices is None:
            indices = np.arange(adata.n_obs)
        if data_loader_class is None:
            data_loader_class = self._data_loader_cls

        if "num_workers" not in data_loader_kwargs:
            data_loader_kwargs.update({"num_workers": settings.dl_num_workers})
        if "persistent_workers" not in data_loader_kwargs:
            data_loader_kwargs.update({"persistent_workers": settings.dl_persistent_workers})

        dl = data_loader_class(
            adata_manager,
            shuffle=shuffle,
            indices=indices,
            batch_size=batch_size,
            **data_loader_kwargs,
        )
        return dl

    def _validate_anndata(
        self, adata: AnnOrMuData | None = None, copy_if_view: bool = True
    ) -> AnnData:
        """Validate anndata has been properly registered, transfer if necessary."""
        if adata is None:
            adata = self.adata

        _check_if_view(adata, copy_if_view=copy_if_view)

        adata_manager = self.get_anndata_manager(adata)
        if adata_manager is None:
            logger.info(
                "Input AnnData not setup with scvi-tools. "
                + "attempting to transfer AnnData setup"
            )
            self._register_manager_for_instance(self.adata_manager.transfer_fields(adata))
        else:
            # Case where correct AnnDataManager is found, replay registration as necessary.
            adata_manager.validate()

        return adata

    def _check_if_trained(self, warn: bool = True, message: str = _UNTRAINED_WARNING_MESSAGE):
        """Check if the model is trained.

        If not trained and `warn` is True, raise a warning, else raise a RuntimeError.
        """
        if not self.is_trained_:
            if warn:
                warnings.warn(message, UserWarning, stacklevel=settings.warnings_stacklevel)
            else:
                raise RuntimeError(message)

    @property
    def is_trained(self) -> bool:
        """Whether the model has been trained."""
        return self.is_trained_

    @property
    def test_indices(self) -> np.ndarray:
        """Observations that are in test set."""
        return self.test_indices_

    @property
    def train_indices(self) -> np.ndarray:
        """Observations that are in train set."""
        return self.train_indices_

    @property
    def validation_indices(self) -> np.ndarray:
        """Observations that are in validation set."""
        return self.validation_indices_

    @train_indices.setter
    def train_indices(self, value):
        self.train_indices_ = value

    @test_indices.setter
    def test_indices(self, value):
        self.test_indices_ = value

    @validation_indices.setter
    def validation_indices(self, value):
        self.validation_indices_ = value

    @is_trained.setter
    def is_trained(self, value):
        self.is_trained_ = value

    @property
    def history(self):
        """Returns computed metrics during training."""
        return self.history_

    def _get_user_attributes(self):
        """Returns all the self attributes defined in a model class, e.g., `self.is_trained_`."""
        attributes = inspect.getmembers(self, lambda a: not (inspect.isroutine(a)))
        attributes = [a for a in attributes if not (a[0].startswith("__") and a[0].endswith("__"))]
        attributes = [a for a in attributes if not a[0].startswith("_abc_")]
        return attributes

    def _get_init_params(self, locals):
        """Returns the model init signature with associated passed in values.

        Ignores the initial AnnData.
        """
        init = self.__init__
        sig = inspect.signature(init)
        parameters = sig.parameters.values()

        init_params = [p.name for p in parameters]
        all_params = {p: locals[p] for p in locals if p in init_params}
        all_params = {
            k: v
            for (k, v) in all_params.items()
            if not isinstance(v, AnnData) and not isinstance(v, MuData)
        }
        # not very efficient but is explicit
        # separates variable params (**kwargs) from non variable params into two dicts
        non_var_params = [p.name for p in parameters if p.kind != p.VAR_KEYWORD]
        non_var_params = {k: v for (k, v) in all_params.items() if k in non_var_params}
        var_params = [p.name for p in parameters if p.kind == p.VAR_KEYWORD]
        var_params = {k: v for (k, v) in all_params.items() if k in var_params}

        user_params = {"kwargs": var_params, "non_kwargs": non_var_params}

        return user_params

    @abstractmethod
    def train(self):
        """Trains the model."""

    def save(
        self,
        dir_path: str,
        prefix: str | None = None,
        overwrite: bool = False,
        save_anndata: bool = False,
        save_kwargs: dict | None = None,
        legacy_mudata_format: bool = False,
        **anndata_write_kwargs,
    ):
        """Save the state of the model.

        Neither the trainer optimizer state nor the trainer history are saved.
        Model files are not expected to be reproducibly saved and loaded across versions
        until we reach version 1.0.

        Parameters
        ----------
        dir_path
            Path to a directory.
        prefix
            Prefix to prepend to saved file names.
        overwrite
            Overwrite existing data or not. If `False` and directory
            already exists at `dir_path`, error will be raised.
        save_anndata
            If True, also saves the anndata
        save_kwargs
            Keyword arguments passed into :func:`~torch.save`.
        legacy_mudata_format
            If ``True``, saves the model ``var_names`` in the legacy format if the model was
            trained with a :class:`~mudata.MuData` object. The legacy format is a flat array with
            variable names across all modalities concatenated, while the new format is a dictionary
            with keys corresponding to the modality names and values corresponding to the variable
            names for each modality.
        anndata_write_kwargs
            Kwargs for :meth:`~anndata.AnnData.write`
        """
        from scvi.model.base._save_load import _get_var_names

        if not os.path.exists(dir_path) or overwrite:
            os.makedirs(dir_path, exist_ok=overwrite)
        else:
            raise ValueError(
                f"{dir_path} already exists. Please provide another directory for saving."
            )

        file_name_prefix = prefix or ""
        save_kwargs = save_kwargs or {}

        if save_anndata:
            file_suffix = ""
            if isinstance(self.adata, AnnData):
                file_suffix = SAVE_KEYS.ADATA_FNAME
            elif isinstance(self.adata, MuData):
                file_suffix = SAVE_KEYS.MDATA_FNAME
            self.adata.write(
                os.path.join(dir_path, f"{file_name_prefix}{file_suffix}"),
                **anndata_write_kwargs,
            )

        model_save_path = os.path.join(dir_path, f"{file_name_prefix}{SAVE_KEYS.MODEL_FNAME}")

        # save the model state dict and the trainer state dict only
        model_state_dict = self.module.state_dict()

        var_names = _get_var_names(self.adata, legacy_mudata_format=legacy_mudata_format)

        # get all the user attributes
        user_attributes = self._get_user_attributes()
        # only save the public attributes with _ at the very end
        user_attributes = {a[0]: a[1] for a in user_attributes if a[0][-1] == "_"}

        torch.save(
            {
                SAVE_KEYS.MODEL_STATE_DICT_KEY: model_state_dict,
                SAVE_KEYS.VAR_NAMES_KEY: var_names,
                SAVE_KEYS.ATTR_DICT_KEY: user_attributes,
            },
            model_save_path,
            **save_kwargs,
        )

    @classmethod
    @devices_dsp.dedent
    def load(
        cls,
        dir_path: str,
        adata: AnnOrMuData | None = None,
        accelerator: str = "auto",
        device: int | str = "auto",
        prefix: str | None = None,
        backup_url: str | None = None,
    ):
        """Instantiate a model from the saved output.

        Parameters
        ----------
        dir_path
            Path to saved outputs.
        adata
            AnnData organized in the same way as data used to train model.
            It is not necessary to run setup_anndata,
            as AnnData is validated against the saved `scvi` setup dictionary.
            If None, will check for and load anndata saved with the model.
        %(param_accelerator)s
        %(param_device)s
        prefix
            Prefix of saved file names.
        backup_url
            URL to retrieve saved outputs from if not present on disk.

        Returns
        -------
        Model with loaded state dictionaries.

        Examples
        --------
        >>> model = ModelClass.load(save_path, adata)
        >>> model.get_....
        """
        load_adata = adata is None
        _, _, device = parse_device_args(
            accelerator=accelerator,
            devices=device,
            return_device="torch",
            validate_single_device=True,
        )

        (
            attr_dict,
            var_names,
            model_state_dict,
            new_adata,
        ) = _load_saved_files(
            dir_path,
            load_adata,
            map_location=device,
            prefix=prefix,
            backup_url=backup_url,
        )
        adata = new_adata if new_adata is not None else adata

        _validate_var_names(adata, var_names)

        registry = attr_dict.pop("registry_")
        if _MODEL_NAME_KEY in registry and registry[_MODEL_NAME_KEY] != cls.__name__:
            raise ValueError("It appears you are loading a model from a different class.")

        if _SETUP_ARGS_KEY not in registry:
            raise ValueError(
                "Saved model does not contain original setup inputs. "
                "Cannot load the original setup."
            )

        # Calling ``setup_anndata`` method with the original arguments passed into
        # the saved model. This enables simple backwards compatibility in the case of
        # newly introduced fields or parameters.
        method_name = registry.get(_SETUP_METHOD_NAME, "setup_anndata")
        getattr(cls, method_name)(adata, source_registry=registry, **registry[_SETUP_ARGS_KEY])

        model = _initialize_model(cls, adata, attr_dict)
        model.module.on_load(model)
        model.module.load_state_dict(model_state_dict)

        model.to_device(device)
        model.module.eval()
        model._validate_anndata(adata)
        return model

    @classmethod
    def convert_legacy_save(
        cls,
        dir_path: str,
        output_dir_path: str,
        overwrite: bool = False,
        prefix: str | None = None,
        **save_kwargs,
    ) -> None:
        """Converts a legacy saved model (<v0.15.0) to the updated save format.

        Parameters
        ----------
        dir_path
            Path to directory where legacy model is saved.
        output_dir_path
            Path to save converted save files.
        overwrite
            Overwrite existing data or not. If ``False`` and directory
            already exists at ``output_dir_path``, error will be raised.
        prefix
            Prefix of saved file names.
        **save_kwargs
            Keyword arguments passed into :func:`~torch.save`.
        """
        if not os.path.exists(output_dir_path) or overwrite:
            os.makedirs(output_dir_path, exist_ok=overwrite)
        else:
            raise ValueError(
                f"{output_dir_path} already exists. Please provide an unexisting directory for "
                "saving."
            )

        file_name_prefix = prefix or ""
        model_state_dict, var_names, attr_dict, _ = _load_legacy_saved_files(
            dir_path, file_name_prefix, load_adata=False
        )

        if "scvi_setup_dict_" in attr_dict:
            scvi_setup_dict = attr_dict.pop("scvi_setup_dict_")
            unlabeled_category_key = "unlabeled_category_"
            unlabeled_category = attr_dict.get(unlabeled_category_key, None)
            attr_dict["registry_"] = registry_from_setup_dict(
                cls,
                scvi_setup_dict,
                unlabeled_category=unlabeled_category,
            )

        model_save_path = os.path.join(output_dir_path, f"{file_name_prefix}model.pt")
        torch.save(
            {
                SAVE_KEYS.MODEL_STATE_DICT_KEY: model_state_dict,
                SAVE_KEYS.VAR_NAMES_KEY: var_names,
                SAVE_KEYS.ATTR_DICT_KEY: attr_dict,
            },
            model_save_path,
            **save_kwargs,
        )

    @property
    def summary_string(self):
        """Summary string of the model."""
        summary_string = self._model_summary_string
        summary_string += "\nTraining status: {}".format(
            "Trained" if self.is_trained_ else "Not Trained"
        )
        return summary_string

    def __repr__(self):
        rich.print(self.summary_string)
        return ""

    @classmethod
    @abstractmethod
    @setup_anndata_dsp.dedent
    def setup_anndata(
        cls,
        adata: AnnData,
        *args,
        **kwargs,
    ):
        """%(summary)s.

        Each model class deriving from this class provides parameters to this method
        according to its needs. To operate correctly with the model initialization,
        the implementation must call :meth:`~scvi.model.base.BaseModelClass.register_manager`
        on a model-specific instance of :class:`~scvi.data.AnnDataManager`.
        """

    @staticmethod
    def view_setup_args(dir_path: str, prefix: str | None = None) -> None:
        """Print args used to setup a saved model.

        Parameters
        ----------
        dir_path
            Path to saved outputs.
        prefix
            Prefix of saved file names.
        """
        registry = BaseModelClass.load_registry(dir_path, prefix)
        AnnDataManager.view_setup_method_args(registry)

    @staticmethod
    def load_registry(dir_path: str, prefix: str | None = None) -> dict:
        """Return the full registry saved with the model.

        Parameters
        ----------
        dir_path
            Path to saved outputs.
        prefix
            Prefix of saved file names.

        Returns
        -------
        The full registry saved with the model
        """
        attr_dict = _load_saved_files(dir_path, False, prefix=prefix)[0]

        # Legacy support for old setup dict format.
        if "scvi_setup_dict_" in attr_dict:
            raise NotImplementedError(
                "Viewing setup args for pre v0.15.0 models is unsupported. "
                "Update your save files with ``convert_legacy_save`` first."
            )

        return attr_dict.pop("registry_")

    def view_anndata_setup(
        self, adata: AnnOrMuData | None = None, hide_state_registries: bool = False
    ) -> None:
        """Print summary of the setup for the initial AnnData or a given AnnData object.

        Parameters
        ----------
        adata
            AnnData object setup with ``setup_anndata`` or
            :meth:`~scvi.data.AnnDataManager.transfer_fields`.
        hide_state_registries
            If True, prints a shortened summary without details of each state registry.
        """
        if adata is None:
            adata = self.adata
        try:
            adata_manager = self.get_anndata_manager(adata, required=True)
        except ValueError as err:
            raise ValueError(
                f"Given AnnData not setup with {self.__class__.__name__}. "
                "Cannot view setup summary."
            ) from err
        adata_manager.view_registry(hide_state_registries=hide_state_registries)


class BaseMinifiedModeModelClass(BaseModelClass):
    """Abstract base class for scvi-tools models that can handle minified data."""

    @property
    def minified_data_type(self) -> MinifiedDataType | None:
        """The type of minified data associated with this model, if applicable."""
        return (
            self.adata_manager.get_from_registry(REGISTRY_KEYS.MINIFY_TYPE_KEY)
            if REGISTRY_KEYS.MINIFY_TYPE_KEY in self.adata_manager.data_registry
            else None
        )

    def minify_adata(
        self,
        minified_data_type: MinifiedDataType = ADATA_MINIFY_TYPE.LATENT_POSTERIOR,
        use_latent_qzm_key: str = "X_latent_qzm",
        use_latent_qzv_key: str = "X_latent_qzv",
    ) -> None:
        """Minify the model's :attr:`~scvi.model.base.BaseModelClass.adata`.

        Minifies the :class:`~anndata.AnnData` object associated with the model according to the
        method specified by ``minified_data_type`` and registers the new fields with the model's
        :class:`~scvi.data.AnnDataManager`. This also sets the ``minified_data_type`` attribute
        of the underlying :class:`~scvi.module.base.BaseModuleClass` instance.

        Parameters
        ----------
        minified_data_type
            Method for minifying the data. One of the following:

            - ``"latent_posterior_parameters"``: Store the latent posterior mean and variance in
                :attr:`~anndata.AnnData.obsm` using the keys ``use_latent_qzm_key`` and
                ``use_latent_qzv_key``.
            - ``"latent_posterior_parameters_with_counts"``: Store the latent posterior mean and
                variance in :attr:`~anndata.AnnData.obsm` using the keys ``use_latent_qzm_key`` and
                ``use_latent_qzv_key``, and the raw count data in :attr:`~anndata.AnnData.X`.
        use_latent_qzm_key
            Key to use for storing the latent posterior mean in :attr:`~anndata.AnnData.obsm` when
            ``minified_data_type`` is ``"latent_posterior"``.
        use_latent_qzv_key
            Key to use for storing the latent posterior variance in :attr:`~anndata.AnnData.obsm`
            when ``minified_data_type`` is ``"latent_posterior"``.

        Notes
        -----
        The modification is not done inplace -- instead the model is assigned a new (minified)
        version of the :class:`~anndata.AnnData`.
        """
        if minified_data_type not in ADATA_MINIFY_TYPE:
            raise NotImplementedError(
                f"Minification method {minified_data_type} is not supported."
            )
        elif not getattr(self.module, "use_observed_lib_size", True):
            raise ValueError(
                "Minification is not supported for models that do not use observed library size."
            )

        keep_count_data = minified_data_type == ADATA_MINIFY_TYPE.LATENT_POSTERIOR_WITH_COUNTS
        mini_adata = get_minified_adata_scrna(
            adata_manager=self.adata_manager,
            keep_count_data=keep_count_data,
        )
        del mini_adata.uns[_SCVI_UUID_KEY]
        mini_adata.uns[_ADATA_MINIFY_TYPE_UNS_KEY] = minified_data_type
        mini_adata.obsm[self._LATENT_QZM_KEY] = self.adata.obsm[use_latent_qzm_key]
        mini_adata.obsm[self._LATENT_QZV_KEY] = self.adata.obsm[use_latent_qzv_key]
        mini_adata.obs[self._OBSERVED_LIB_SIZE_KEY] = np.squeeze(
            np.asarray(self.adata_manager.get_from_registry(REGISTRY_KEYS.X_KEY).sum(axis=-1))
        )
        self._update_adata_and_manager_post_minification(
            mini_adata,
            minified_data_type,
        )
        self.module.minified_data_type = minified_data_type

    @classmethod
    def _get_fields_for_adata_minification(
        cls,
        minified_data_type: MinifiedDataType,
    ):
        """Return the fields required for minification of the given type."""
        if minified_data_type not in ADATA_MINIFY_TYPE:
            raise NotImplementedError(
                f"Minification method {minified_data_type} is not supported."
            )

        mini_fields = [
            fields.ObsmField(REGISTRY_KEYS.LATENT_QZM_KEY, cls._LATENT_QZM_KEY),
            fields.ObsmField(REGISTRY_KEYS.LATENT_QZV_KEY, cls._LATENT_QZV_KEY),
            fields.NumericalObsField(REGISTRY_KEYS.OBSERVED_LIB_SIZE, cls._OBSERVED_LIB_SIZE_KEY),
            fields.StringUnsField(REGISTRY_KEYS.MINIFY_TYPE_KEY, _ADATA_MINIFY_TYPE_UNS_KEY),
        ]

        return mini_fields

    def _update_adata_and_manager_post_minification(
        self,
        minified_adata: AnnOrMuData,
        minified_data_type: MinifiedDataType,
    ):
        """Update the :class:`~anndata.AnnData` and :class:`~scvi.data.AnnDataManager` in-place.

        Parameters
        ----------
        minified_adata
            Minified version of :attr:`~scvi.model.base.BaseModelClass.adata`.
        minified_data_type
            Method used for minifying the data.
        keep_count_data
            If ``True``, the full count matrix is kept in the minified
            :attr:`~scvi.model.base.BaseModelClass.adata`.
        """
        self._validate_anndata(minified_adata)
        new_adata_manager = self.get_anndata_manager(minified_adata, required=True)
        new_adata_manager.register_new_fields(
            self._get_fields_for_adata_minification(minified_data_type)
        )
        self.adata = minified_adata

    @property
    def summary_string(self):
        """Summary string of the model."""
        summary_string = super().summary_string
        summary_string += "\nModel's adata is minified?: {}".format(
            hasattr(self, "minified_data_type") and self.minified_data_type is not None
        )
        return summary_string


class BaseMudataMinifiedModeModelClass(BaseModelClass):
    """Abstract base class for scvi-tools models that can handle minified data."""

    @property
    def minified_data_type(self) -> MinifiedDataType | None:
        """The type of minified data associated with this model, if applicable."""
        return (
            self.adata_manager.get_from_registry(REGISTRY_KEYS.MINIFY_TYPE_KEY)
            if REGISTRY_KEYS.MINIFY_TYPE_KEY in self.adata_manager.data_registry
            else None
        )

    def minify_mudata(
        self,
        minified_data_type: MinifiedDataType = ADATA_MINIFY_TYPE.LATENT_POSTERIOR,
        use_latent_qzm_key: str = "X_latent_qzm",
        use_latent_qzv_key: str = "X_latent_qzv",
    ) -> None:
        """Minify the model's :attr:`~scvi.model.base.BaseModelClass.adata`.

        Minifies the :class:`~mudata.MuData` object associated with the model according to the
        method specified by ``minified_data_type`` and registers the new fields with the model's
        :class:`~scvi.data.AnnDataManager`. This also sets the ``minified_data_type`` attribute
        of the underlying :class:`~scvi.module.base.BaseModuleClass` instance.

        Parameters
        ----------
        minified_data_type
            Method for minifying the data. One of the following:

            - ``"latent_posterior_parameters"``: Store the latent posterior mean and variance in
                :attr:`~mudata.MuData.obsm` using the keys ``use_latent_qzm_key`` and
                ``use_latent_qzv_key``.
            - ``"latent_posterior_parameters_with_counts"``: Store the latent posterior mean and
                variance in :attr:`~mudata.MuData.obsm` using the keys ``use_latent_qzm_key`` and
                ``use_latent_qzv_key``, and the raw count data in :attr:`~mudata.MuData.X`.
        use_latent_qzm_key
            Key to use for storing the latent posterior mean in :attr:`~mudata.MuData.obsm` when
            ``minified_data_type`` is ``"latent_posterior"``.
        use_latent_qzv_key
            Key to use for storing the latent posterior variance in :attr:`~mudata.MuData.obsm`
            when ``minified_data_type`` is ``"latent_posterior"``.

        Notes
        -----
        The modification is not done inplace -- instead the model is assigned a new (minified)
        version of the :class:`~mudata.MuData`.
        """
        if self.adata_manager._registry["setup_method_name"] != "setup_mudata":
            raise ValueError(
                f"MuData must be registered with {self.__name__}.setup_mudata to use this method."
            )
        if minified_data_type not in ADATA_MINIFY_TYPE:
            raise NotImplementedError(
                f"Minification method {minified_data_type} is not supported."
            )
        elif not getattr(self.module, "use_observed_lib_size", True):
            raise ValueError(
                "Minification is not supported for models that do not use observed library size."
            )

        keep_count_data = minified_data_type == ADATA_MINIFY_TYPE.LATENT_POSTERIOR_WITH_COUNTS
        mini_adata = get_minified_mudata(
            adata_manager=self.adata_manager,
            keep_count_data=keep_count_data,
        )
        del mini_adata.uns[_SCVI_UUID_KEY]
        mini_adata.uns[_ADATA_MINIFY_TYPE_UNS_KEY] = minified_data_type
        mini_adata.obsm[self._LATENT_QZM_KEY] = self.adata.obsm[use_latent_qzm_key]
        mini_adata.obsm[self._LATENT_QZV_KEY] = self.adata.obsm[use_latent_qzv_key]
        mini_adata.obs[self._OBSERVED_LIB_SIZE_KEY] = np.squeeze(
            np.asarray(self.adata_manager.get_from_registry(REGISTRY_KEYS.X_KEY).sum(axis=-1))
        )
        self._update_mudata_and_manager_post_minification(
            mini_adata,
            minified_data_type,
        )
        self.module.minified_data_type = minified_data_type

    @classmethod
    def _get_fields_for_mudata_minification(
        cls,
        minified_data_type: MinifiedDataType,
    ):
        """Return the fields required for minification of the given type."""
        if minified_data_type not in ADATA_MINIFY_TYPE:
            raise NotImplementedError(
                f"Minification method {minified_data_type} is not supported."
            )

        mini_fields = [
            fields.ObsmField(REGISTRY_KEYS.LATENT_QZM_KEY, cls._LATENT_QZM_KEY),
            fields.ObsmField(REGISTRY_KEYS.LATENT_QZV_KEY, cls._LATENT_QZV_KEY),
            fields.NumericalObsField(REGISTRY_KEYS.OBSERVED_LIB_SIZE, cls._OBSERVED_LIB_SIZE_KEY),
            fields.StringUnsField(REGISTRY_KEYS.MINIFY_TYPE_KEY, _ADATA_MINIFY_TYPE_UNS_KEY),
        ]

        return mini_fields

    def _update_mudata_and_manager_post_minification(
        self, minified_adata: AnnOrMuData, minified_data_type: MinifiedDataType
    ):
        """Update the mudata and manager inplace after creating a minified adata."""
        # Register this new adata with the model, creating a new manager in the cache
        self._validate_anndata(minified_adata)
        new_adata_manager = self.get_anndata_manager(minified_adata, required=True)
        # This inplace edits the manager
        new_adata_manager.register_new_fields(
            self._get_fields_for_mudata_minification(minified_data_type),
        )
        new_adata_manager.registry["setup_method_name"] = "setup_mudata"
        # We set the adata attribute of the model as this will update self.registry_
        # and self.adata_manager with the new adata manager
        self.adata = minified_adata

    @property
    def summary_string(self):
        """Summary string of the model."""
        summary_string = super().summary_string
        summary_string += "\nModel's adata is minified?: {}".format(
            hasattr(self, "minified_data_type") and self.minified_data_type is not None
        )
        return summary_string



from functools import partial

import numpy as np
import pytest

from scvi.data import synthetic_iid
from scvi.model import SCVI
from scvi.model.base._de_core import _prepare_obs
from scvi.model.base._differential import (
    DifferentialComputation,
    estimate_delta,
    estimate_pseudocounts_offset,
)


def test_features():
    a = np.random.randn(
        100,
    )
    b = 3 + np.random.randn(
        100,
    )
    c = -3 + np.random.randn(
        100,
    )
    alls = np.concatenate([a, b, c])
    delta = estimate_delta(alls)
    expected_range = (delta >= 0.4 * 3) and (delta <= 6)
    if not expected_range:
        raise ValueError("The effect-size threshold was not properly estimated.")

    scales_a = np.random.rand(100, 50)
    where_zero_a = np.zeros(50, dtype=bool)
    where_zero_a[:10] = True
    scales_a[:, :10] = 1e-6

    scales_b = np.random.rand(100, 50)
    where_zero_b = np.zeros(50, dtype=bool)
    where_zero_b[-10:] = True
    scales_b[:, -10:] = 1e-7
    offset = estimate_pseudocounts_offset(
        scales_a=scales_a,
        scales_b=scales_b,
        where_zero_a=where_zero_a,
        where_zero_b=where_zero_b,
    )
    expected_off_range = offset <= 1e-6
    if not expected_off_range:
        raise ValueError("The pseudocount offset was not properly estimated.")


def test_differential_computation(save_path):
    n_latent = 5
    adata = synthetic_iid(batch_size=50, n_genes=30, n_proteins=20, n_regions=20)
    SCVI.setup_anndata(
        adata,
        batch_key="batch",
        labels_key="labels",
    )
    model = SCVI(adata, n_latent=n_latent)
    model.train(1)

    model_fn = partial(model.get_normalized_expression, return_numpy=True)
    rep_fn = model.get_latent_representation
    dc = DifferentialComputation(model_fn, rep_fn, model.adata_manager)
    cell_idx1 = np.asarray(adata.obs.labels == "label_1")
    cell_idx2 = ~cell_idx1
    dc.get_bayes_factors(cell_idx1, cell_idx2, mode="vanilla", use_permutation=True)
    res = dc.get_bayes_factors(cell_idx1, cell_idx2, mode="change", use_permutation=False)

    model_fn = partial(
        model.get_normalized_expression,
        weights="importance",
        return_numpy=True,
    )
    dc = DifferentialComputation(
        model_fn,
        rep_fn,
        model.adata_manager,
    )
    cell_idx1 = np.asarray(adata.obs.labels == "label_1")
    cell_idx2 = ~cell_idx1
    dc.get_bayes_factors(cell_idx1, cell_idx2, mode="vanilla", use_permutation=True)
    res = dc.get_bayes_factors(cell_idx1, cell_idx2, mode="change", use_permutation=False)
    assert res["delta"] == 0.5
    assert res["pseudocounts"] == 0.0
    res = dc.get_bayes_factors(
        cell_idx1, cell_idx2, mode="change", use_permutation=False, delta=None
    )
    dc.get_bayes_factors(
        cell_idx1,
        cell_idx2,
        mode="change",
        use_permutation=False,
        delta=None,
        pseudocounts=None,
    )
    dc.get_bayes_factors(cell_idx1, cell_idx2, mode="change", cred_interval_lvls=[0.75])

    delta = 0.5

    def change_fn_test(x, y):
        return x - y

    def m1_domain_fn_test(samples):
        return np.abs(samples) >= delta

    dc.get_bayes_factors(
        cell_idx1,
        cell_idx2,
        mode="change",
        m1_domain_fn=m1_domain_fn_test,
        change_fn=change_fn_test,
    )

    # should fail if just one batch
    with pytest.raises(ValueError):
        model.differential_expression(adata[:20], groupby="batch")

    # test view
    model.differential_expression(adata[adata.obs["labels"] == "label_1"], groupby="batch")

    # Test query features
    (
        obs_col,
        group1,
        _,
    ) = _prepare_obs(idx1="(labels == 'label_1') & (batch == 'batch_1')", idx2=None, adata=adata)
    assert (obs_col == group1).sum() == adata.obs.loc[
        lambda x: (x.labels == "label_1") & (x.batch == "batch_1")
    ].shape[0]
    model.differential_expression(
        idx1="labels == 'label_1'",
    )
    model.differential_expression(
        idx1="labels == 'label_1'", idx2="(labels == 'label_2') & (batch == 'batch_1')"
    )

    # test that ints as group work
    a = synthetic_iid()
    SCVI.setup_anndata(
        a,
        batch_key="batch",
        labels_key="labels",
    )
    a.obs["test"] = [0] * 200 + [1] * 200
    model = SCVI(a)
    model.train(1)
    model.differential_expression(groupby="test", group1=0)

    # test that string but not as categorical work
    a = synthetic_iid()
    SCVI.setup_anndata(
        a,
        batch_key="batch",
        labels_key="labels",
    )
    a.obs["test"] = ["0"] * 200 + ["1"] * 200
    model = SCVI(a)
    model.train(1)
    model.differential_expression(groupby="test", group1="0")



from __future__ import annotations

from typing import TYPE_CHECKING

import torch

from scvi import REGISTRY_KEYS
from scvi.module._constants import MODULE_KEYS
from scvi.module.base import BaseModuleClass, auto_move_data

if TYPE_CHECKING:
    import numpy as np
    from torch.distributions import Distribution


class VAEC(BaseModuleClass):
    """Conditional Variational auto-encoder model.

    This is an implementation of the CondSCVI model

    Parameters
    ----------
    n_input
        Number of input genes
    n_batch
        Number of batches. If ``0``, no batch correction is performed.
    n_labels
        Number of labels
    n_hidden
        Number of nodes per hidden layer
    n_latent
        Dimensionality of the latent space
    n_layers
        Number of hidden layers used for encoder and decoder NNs
    log_variational
        Log(data+1) prior to encoding for numerical stability. Not normalization.
    ct_weight
        Multiplicative weight for cell type specific latent space.
    dropout_rate
        Dropout rate for the encoder and decoder neural network.
    encode_covariates
        If ``True``, covariates are concatenated to gene expression prior to passing through
        the encoder(s). Else, only gene expression is used.
    extra_encoder_kwargs
        Keyword arguments passed into :class:`~scvi.nn.Encoder`.
    extra_decoder_kwargs
        Keyword arguments passed into :class:`~scvi.nn.FCLayers`.
    """

    def __init__(
        self,
        n_input: int,
        n_batch: int = 0,
        n_labels: int = 0,
        n_hidden: int = 128,
        n_latent: int = 5,
        n_layers: int = 2,
        log_variational: bool = True,
        ct_weight: np.ndarray | None = None,
        dropout_rate: float = 0.05,
        encode_covariates: bool = False,
        extra_encoder_kwargs: dict | None = None,
        extra_decoder_kwargs: dict | None = None,
    ):
        from scvi.nn import Encoder, FCLayers

        super().__init__()
        self.dispersion = "gene"
        self.n_latent = n_latent
        self.n_layers = n_layers
        self.n_hidden = n_hidden
        self.dropout_rate = dropout_rate
        self.log_variational = log_variational
        self.gene_likelihood = "nb"
        self.latent_distribution = "normal"
        self.encode_covariates = encode_covariates
        self.n_batch = n_batch
        self.n_labels = n_labels

        if self.encode_covariates and self.n_batch < 1:
            raise ValueError("`n_batch` must be greater than 0 if `encode_covariates` is `True`.")

        self.px_r = torch.nn.Parameter(torch.randn(n_input))
        self.z_encoder = Encoder(
            n_input,
            n_latent,
            n_cat_list=[n_labels] + ([n_batch] if n_batch > 0 and encode_covariates else []),
            n_layers=n_layers,
            n_hidden=n_hidden,
            dropout_rate=dropout_rate,
            inject_covariates=True,
            use_batch_norm=False,
            use_layer_norm=True,
            return_dist=True,
            **(extra_encoder_kwargs or {}),
        )

        self.decoder = FCLayers(
            n_in=n_latent,
            n_out=n_hidden,
            n_cat_list=[n_labels] + ([n_batch] if n_batch > 0 else []),
            n_layers=n_layers,
            n_hidden=n_hidden,
            dropout_rate=dropout_rate,
            inject_covariates=True,
            use_batch_norm=False,
            use_layer_norm=True,
            **(extra_decoder_kwargs or {}),
        )
        self.px_decoder = torch.nn.Sequential(
            torch.nn.Linear(n_hidden, n_input), torch.nn.Softplus()
        )

        self.register_buffer(
            "ct_weight",
            (
                torch.ones((self.n_labels,), dtype=torch.float32)
                if ct_weight is None
                else torch.tensor(ct_weight, dtype=torch.float32)
            ),
        )

    def _get_inference_input(
        self, tensors: dict[str, torch.Tensor]
    ) -> dict[str, torch.Tensor | None]:
        return {
            MODULE_KEYS.X_KEY: tensors[REGISTRY_KEYS.X_KEY],
            MODULE_KEYS.Y_KEY: tensors[REGISTRY_KEYS.LABELS_KEY],
            MODULE_KEYS.BATCH_INDEX_KEY: tensors.get(REGISTRY_KEYS.BATCH_KEY, None),
        }

    def _get_generative_input(
        self,
        tensors: dict[str, torch.Tensor],
        inference_outputs: dict[str, torch.Tensor | Distribution],
    ) -> dict[str, torch.Tensor]:
        return {
            MODULE_KEYS.Z_KEY: inference_outputs[MODULE_KEYS.Z_KEY],
            MODULE_KEYS.LIBRARY_KEY: inference_outputs[MODULE_KEYS.LIBRARY_KEY],
            MODULE_KEYS.Y_KEY: tensors[REGISTRY_KEYS.LABELS_KEY],
            MODULE_KEYS.BATCH_INDEX_KEY: tensors.get(REGISTRY_KEYS.BATCH_KEY, None),
        }

    @auto_move_data
    def inference(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        batch_index: torch.Tensor | None = None,
        n_samples: int = 1,
    ) -> dict[str, torch.Tensor | Distribution]:
        """High level inference method.

        Runs the inference (encoder) model.
        """
        x_ = x
        library = x.sum(1).unsqueeze(1)
        if self.log_variational:
            x_ = torch.log1p(x_)

        encoder_input = [x, y]
        if batch_index is not None and self.encode_covariates:
            encoder_input.append(batch_index)

        qz, z = self.z_encoder(*encoder_input)

        if n_samples > 1:
            untran_z = qz.sample((n_samples,))
            z = self.z_encoder.z_transformation(untran_z)
            library = library.unsqueeze(0).expand((n_samples, library.size(0), library.size(1)))

        return {
            MODULE_KEYS.Z_KEY: z,
            MODULE_KEYS.QZ_KEY: qz,
            MODULE_KEYS.LIBRARY_KEY: library,
        }

    @auto_move_data
    def generative(
        self,
        z: torch.Tensor,
        library: torch.Tensor,
        y: torch.Tensor,
        batch_index: torch.Tensor | None = None,
        transform_batch: torch.Tensor | None = None,
    ) -> dict[str, Distribution]:
        """Runs the generative model."""
        from scvi.distributions import NegativeBinomial

        decoder_input = [z, y]
        if transform_batch is not None:
            batch_index = torch.ones_like(batch_index) * transform_batch

        if batch_index is not None:
            decoder_input.append(batch_index)

        h = self.decoder(*decoder_input)
        px_scale = self.px_decoder(h)
        px_rate = library * px_scale
        return {MODULE_KEYS.PX_KEY: NegativeBinomial(px_rate, logits=self.px_r)}

    def loss(
        self,
        tensors: dict[str, torch.Tensor],
        inference_outputs: dict[str, torch.Tensor | Distribution],
        generative_outputs: dict[str, Distribution],
        kl_weight: float = 1.0,
    ):
        """Loss computation."""
        from torch.distributions import Normal
        from torch.distributions import kl_divergence as kl

        from scvi.module.base import LossOutput

        x = tensors[REGISTRY_KEYS.X_KEY]
        y = tensors[REGISTRY_KEYS.LABELS_KEY]
        qz = inference_outputs[MODULE_KEYS.QZ_KEY]
        px = generative_outputs[MODULE_KEYS.PX_KEY]

        mean = torch.zeros_like(qz.loc)
        scale = torch.ones_like(qz.scale)

        kl_divergence_z = kl(qz, Normal(mean, scale)).sum(dim=1)

        reconst_loss = -px.log_prob(x).sum(-1)
        scaling_factor = self.ct_weight[y.long()[:, 0]]
        loss = torch.mean(scaling_factor * (reconst_loss + kl_weight * kl_divergence_z))

        return LossOutput(loss=loss, reconstruction_loss=reconst_loss, kl_local=kl_divergence_z)

    @torch.inference_mode()
    def sample(
        self,
        tensors: dict[str, torch.Tensor],
        n_samples: int = 1,
    ) -> torch.Tensor:
        r"""Generate observation samples from the posterior predictive distribution.

        The posterior predictive distribution is written as :math:`p(\hat{x} \mid x)`.

        Parameters
        ----------
        tensors
            Tensors dict
        n_samples
            Number of required samples for each cell

        Returns
        -------
        x_new : :py:class:`torch.Tensor`
            tensor with shape (n_cells, n_genes, n_samples)
        """
        inference_kwargs = {"n_samples": n_samples}
        generative_outputs = self.forward(
            tensors,
            inference_kwargs=inference_kwargs,
            compute_loss=False,
        )[1]

        dist = generative_outputs[MODULE_KEYS.PX_KEY]
        if n_samples > 1:
            exprs = dist.sample().permute([1, 2, 0])  # Shape : (n_cells_batch, n_genes, n_samples)
        else:
            exprs = dist.sample()

        return exprs.cpu()



from typing import NamedTuple


class _METRIC_KEYS_NT(NamedTuple):
    TRAINING_KEY: str = "training"
    VALIDATION_KEY: str = "validation"
    # classification
    ACCURACY_KEY: str = "accuracy"
    F1_SCORE_KEY: str = "f1_score"
    CALIBRATION_ERROR_KEY: str = "calibration_error"
    AUROC_KEY: str = "auroc"
    CLASSIFICATION_LOSS_KEY: str = "classification_loss"
    TRUE_LABELS_KEY: str = "true_labels"
    LOGITS_KEY: str = "logits"


METRIC_KEYS = _METRIC_KEYS_NT()



import os

import pytest

from scvi.data import synthetic_iid
from scvi.model import CondSCVI


@pytest.mark.parametrize("n_batches", [1, 2, 3])
@pytest.mark.parametrize("encode_covariates", [True, False])
def test_condscvi_batch_key(
    save_path: str, n_batches: int, encode_covariates: bool, n_labels: int = 5
):
    adata = synthetic_iid(n_batches=n_batches, n_labels=n_labels)
    CondSCVI.setup_anndata(adata, batch_key="batch", labels_key="labels")
    model = CondSCVI(adata, encode_covariates=encode_covariates)

    model.train(max_epochs=1)
    assert model.summary_stats.n_batch == n_batches
    _ = model.get_elbo()
    _ = model.get_reconstruction_error()
    _ = model.get_latent_representation()
    _ = model.get_vamp_prior(adata)

    model.get_normalized_expression(adata)
    model.differential_expression(groupby="labels", group1="label_1")
    model.differential_expression(groupby="labels", group1="label_1", group2="label_2")

    model_path = os.path.join(save_path, __name__)
    model.save(model_path, overwrite=True, save_anndata=False)
    model = CondSCVI.load(model_path, adata=adata)


def test_condscvi_batch_key_compat_load(save_path: str):
    adata = synthetic_iid()
    model = CondSCVI.load("tests/test_data/condscvi_pre_batch", adata=adata)

    assert not hasattr(model.summary_stats, "n_batch")
    _ = model.get_latent_representation()
    _ = model.get_vamp_prior(adata)

    model_path = os.path.join(save_path, __name__)
    model.save(model_path, overwrite=True, save_anndata=False)
    model = CondSCVI.load(model_path, adata=adata)


@pytest.mark.parametrize("weight_obs", [True, False])
def test_condscvi_no_batch_key(save_path: str, weight_obs: bool):
    adata = synthetic_iid()
    CondSCVI.setup_anndata(adata, labels_key="labels")

    with pytest.raises(ValueError):
        _ = CondSCVI(adata, encode_covariates=True)

    model = CondSCVI(adata, weight_obs=weight_obs)
    model.train(max_epochs=1)
    assert not hasattr(model.summary_stats, "n_batch")
    _ = model.get_elbo()
    _ = model.get_reconstruction_error()
    _ = model.get_latent_representation()
    _ = model.get_vamp_prior(adata)

    model_path = os.path.join(save_path, __name__)
    model.save(model_path, overwrite=True, save_anndata=False)
    model = CondSCVI.load(model_path, adata=adata)



import jax
import jax.numpy as jnp

from scvi.external.mrvi._components import (
    AttentionBlock,
    ConditionalNormalization,
    Dense,
    NormalDistOutputNN,
    ResnetBlock,
)


def test_dense():
    key = jax.random.PRNGKey(0)
    x = jnp.ones((20, 10))
    dense = Dense(10)
    params = dense.init(key, x)
    dense.apply(params, x)


def test_resnetblock():
    key = jax.random.PRNGKey(0)
    x = jnp.ones((20, 10))
    block = ResnetBlock(10, 30, training=True)
    params = block.init(key, x)
    block.apply(params, x, mutable=["batch_stats"])


def test_normalnn():
    key = jax.random.PRNGKey(0)
    key, subkey = jax.random.split(key)
    x = jnp.ones((20, 10))
    nn = NormalDistOutputNN(10, 30, 3, training=True)
    params = nn.init(key, x)
    nn.apply(params, x, mutable=["batch_stats"])


def test_conditionalbatchnorm1d():
    key = jax.random.PRNGKey(0)
    x = jnp.ones((20, 10))
    y = jnp.ones((20, 1))
    conditionalbatchnorm1d = ConditionalNormalization(
        10, 3, normalization_type="batch", training=True
    )
    params = conditionalbatchnorm1d.init(key, x, y)
    conditionalbatchnorm1d.apply(params, x, y, mutable=["batch_stats"])


def test_attention():
    key = jax.random.PRNGKey(0)
    q_vals = jnp.ones((30, 10))
    kv_vals = jnp.ones((30, 10))
    mod = AttentionBlock(query_dim=20, out_dim=40, training=True)
    params = mod.init(key, q_vals, kv_vals)
    mod.apply(params, q_vals, kv_vals, mutable=["batch_stats"])
    q_vals_3d = jnp.ones((3, 30, 10))
    kv_vals_3d = jnp.ones((3, 30, 10))
    r = mod.apply(params, q_vals_3d, kv_vals_3d, mutable=["batch_stats"])
    assert r[0].shape == (3, 30, 40)



from __future__ import annotations

import logging
import warnings
from typing import TYPE_CHECKING

import jax
import jax.numpy as jnp
import numpy as np
import xarray as xr
from tqdm import tqdm

from scvi import REGISTRY_KEYS
from scvi.data import AnnDataManager, fields
from scvi.external.mrvi._module import MRVAE
from scvi.external.mrvi._types import MRVIReduction
from scvi.external.mrvi._utils import rowwise_max_excluding_diagonal
from scvi.model.base import BaseModelClass, JaxTrainingMixin
from scvi.utils import setup_anndata_dsp
from scvi.utils._docstrings import devices_dsp

if TYPE_CHECKING:
    from typing import Literal

    import numpy.typing as npt
    from anndata import AnnData
    from numpyro.distributions import Distribution

logger = logging.getLogger(__name__)

DEFAULT_TRAIN_KWARGS = {
    "max_epochs": 100,
    "early_stopping": True,
    "early_stopping_patience": 15,
    "check_val_every_n_epoch": 1,
    "batch_size": 256,
    "train_size": 0.9,
    "plan_kwargs": {
        "lr": 2e-3,
        "n_epochs_kl_warmup": 20,
        "max_norm": 40,
        "eps": 1e-8,
        "weight_decay": 1e-8,
    },
}


class MRVI(JaxTrainingMixin, BaseModelClass):
    """Multi-resolution Variational Inference (MrVI) :cite:p:`Boyeau24`.

    Parameters
    ----------
    adata
        AnnData object that has been registered via :meth:`~scvi.external.MRVI.setup_anndata`.
    n_latent
        Dimensionality of the latent space for ``z``.
    n_latent_u
        Dimensionality of the latent space for ``u``.
    encoder_n_hidden
        Number of nodes per hidden layer in the encoder.
    encoder_n_layers
        Number of hidden layers in the encoder.
    z_u_prior
        Whether to use a prior for ``z_u``.
    z_u_prior_scale
        Scale of the prior for the difference between ``z`` and ``u``.
    u_prior_scale
        Scale of the prior for ``u``.
    u_prior_mixture
        Whether to use a mixture model for the ``u`` prior.
    u_prior_mixture_k
        Number of components in the mixture model for the ``u`` prior.
    learn_z_u_prior_scale
        Whether to learn the scale of the ``z`` and ``u`` difference prior during training.
    laplace_scale
        Scale parameter for the Laplace distribution in the decoder.
    scale_observations
        Whether to scale loss by the number of observations per sample.
    px_kwargs
        Keyword args for :class:`~scvi.external.mrvi._module.DecoderZXAttention`.
    qz_kwargs
        Keyword args for :class:`~scvi.external.mrvi._module.EncoderUZ`.
    qu_kwargs
        Keyword args for :class:`~scvi.external.mrvi._module.EncoderXU`.

    Notes
    -----
    See further usage examples in the following tutorial:

    1. :doc:`/tutorials/notebooks/scrna/MrVI_tutorial`

    See the user guide for this model:

    1. :doc:`/user_guide/models/mrvi`

    See Also
    --------
    :class:`~scvi.external.mrvi.MRVAE`
    """

    def __init__(self, adata: AnnData, **model_kwargs):
        super().__init__(adata)

        n_sample = self.summary_stats.n_sample
        n_batch = self.summary_stats.n_batch
        n_labels = self.summary_stats.n_labels

        obs_df = adata.obs.copy()
        obs_df = obs_df.loc[~obs_df._scvi_sample.duplicated("first")]
        self.sample_info = obs_df.set_index("_scvi_sample").sort_index()
        self.sample_key = self.adata_manager.get_state_registry(
            REGISTRY_KEYS.SAMPLE_KEY
        ).original_key
        self.sample_order = self.adata_manager.get_state_registry(
            REGISTRY_KEYS.SAMPLE_KEY
        ).categorical_mapping

        self.n_obs_per_sample = jnp.array(
            adata.obs._scvi_sample.value_counts().sort_index().values
        )

        self.module = MRVAE(
            n_input=self.summary_stats.n_vars,
            n_sample=n_sample,
            n_batch=n_batch,
            n_labels=n_labels,
            n_obs_per_sample=self.n_obs_per_sample,
            **model_kwargs,
        )
        self.init_params_ = self._get_init_params(locals())

    def to_device(self, device):
        # TODO(jhong): remove this once we have a better way to handle device.
        pass

    def _generate_stacked_rngs(self, n_sets: int | tuple) -> dict[str, jax.random.KeyArray]:
        return_1d = isinstance(n_sets, int)
        if return_1d:
            n_sets_1d = n_sets
        else:
            n_sets_1d = np.prod(n_sets)
        rngs_list = [self.module.rngs for _ in range(n_sets_1d)]
        # Combine list of RNG dicts into a single list. This is necessary for vmap/map.
        rngs = {
            required_rng: jnp.concatenate(
                [rngs_dict[required_rng][None] for rngs_dict in rngs_list], axis=0
            )
            for required_rng in self.module.required_rngs
        }
        if not return_1d:
            # Reshaping the random keys to the desired shape in
            # the case of multiple sets.
            rngs = {
                key: random_key.reshape(n_sets + random_key.shape[1:])
                for (key, random_key) in rngs.items()
            }
        return rngs

    @classmethod
    @setup_anndata_dsp.dedent
    def setup_anndata(
        cls,
        adata: AnnData,
        layer: str | None = None,
        sample_key: str | None = None,
        batch_key: str | None = None,
        labels_key: str | None = None,
        **kwargs,
    ):
        """%(summary)s.

        Parameters
        ----------
        %(param_adata)s
        %(param_layer)s
        %(param_sample_key)s
        %(param_batch_key)s
        %(param_labels_key)s
        **kwargs
            Additional keyword arguments passed into
            :meth:`~scvi.data.AnnDataManager.register_fields`.
        """
        setup_method_args = cls._get_setup_method_args(**locals())
        # Add index for batched computation of local statistics.
        adata.obs["_indices"] = np.arange(adata.n_obs).astype(int)
        anndata_fields = [
            fields.LayerField(REGISTRY_KEYS.X_KEY, layer, is_count_data=True),
            fields.CategoricalObsField(REGISTRY_KEYS.SAMPLE_KEY, sample_key),
            fields.CategoricalObsField(REGISTRY_KEYS.BATCH_KEY, batch_key),
            fields.CategoricalObsField(REGISTRY_KEYS.LABELS_KEY, labels_key),
            fields.NumericalObsField(REGISTRY_KEYS.INDICES_KEY, "_indices"),
        ]

        adata_manager = AnnDataManager(fields=anndata_fields, setup_method_args=setup_method_args)
        adata_manager.register_fields(adata, **kwargs)
        cls.register_manager(adata_manager)

    @devices_dsp.dedent
    def train(
        self,
        max_epochs: int | None = None,
        accelerator: str | None = "auto",
        devices: int | list[int] | str = "auto",
        train_size: float | None = None,
        validation_size: float | None = None,
        batch_size: int = 128,
        early_stopping: bool = False,
        plan_kwargs: dict | None = None,
        **trainer_kwargs,
    ):
        """Train the model.

        Parameters
        ----------
        max_epochs
            Maximum number of epochs to train the model. The actual number of epochs may be less if
            early stopping is enabled. If ``None``, defaults to a heuristic based on
            :func:`~scvi.model.get_max_epochs_heuristic`.
        %(param_accelerator)s
        %(param_devices)s
        train_size
            Size of training set in the range ``[0.0, 1.0]``.
        validation_size
            Size of the validation set. If ``None``, defaults to ``1 - train_size``. If
            ``train_size + validation_size < 1``, the remaining cells belong to a test set.
        batch_size
            Minibatch size to use during training.
        early_stopping
            Perform early stopping. Additional arguments can be passed in through ``**kwargs``.
            See :class:`~scvi.train.Trainer` for further options.
        plan_kwargs
            Additional keyword arguments passed into :class:`~scvi.train.JaxTrainingPlan`.
        **trainer_kwargs
            Additional keyword arguments passed into :class:`~scvi.train.Trainer`.
        """
        from copy import deepcopy

        train_kwargs = {
            "max_epochs": max_epochs,
            "accelerator": accelerator,
            "devices": devices,
            "train_size": train_size,
            "validation_size": validation_size,
            "batch_size": batch_size,
            "early_stopping": early_stopping,
            **trainer_kwargs,
        }
        train_kwargs = dict(deepcopy(DEFAULT_TRAIN_KWARGS), **train_kwargs)
        plan_kwargs = plan_kwargs or {}
        train_kwargs["plan_kwargs"] = dict(
            deepcopy(DEFAULT_TRAIN_KWARGS["plan_kwargs"]), **plan_kwargs
        )
        super().train(**train_kwargs)

    def get_latent_representation(
        self,
        adata: AnnData | None = None,
        indices: npt.ArrayLike | None = None,
        batch_size: int | None = None,
        use_mean: bool = True,
        give_z: bool = False,
    ) -> npt.NDArray:
        """Compute the latent representation of the data.

        Parameters
        ----------
        adata
            AnnData object to use. Defaults to the AnnData object used to initialize the model.
        indices
            Indices of cells to use.
        batch_size
            Batch size to use for computing the latent representation.
        use_mean
            Whether to use the mean of the distribution as the latent representation.
        give_z
            Whether to return the z latent representation or the u latent representation.

        Returns
        -------
        The latent representation of the data.
        """
        self._check_if_trained(warn=False)
        adata = self._validate_anndata(adata)
        scdl = self._make_data_loader(
            adata=adata, indices=indices, batch_size=batch_size, iter_ndarray=True
        )

        us = []
        zs = []
        jit_inference_fn = self.module.get_jit_inference_fn(
            inference_kwargs={"use_mean": use_mean}
        )
        for array_dict in tqdm(scdl):
            outputs = jit_inference_fn(self.module.rngs, array_dict)

            if give_z:
                zs.append(jax.device_get(outputs["z"]))
            else:
                us.append(jax.device_get(outputs["u"]))

        if give_z:
            return np.array(jnp.concatenate(zs, axis=0))
        else:
            return np.array(jnp.concatenate(us, axis=0))

    def compute_local_statistics(
        self,
        reductions: list[MRVIReduction],
        adata: AnnData | None = None,
        indices: npt.ArrayLike | None = None,
        batch_size: int | None = None,
        use_vmap: bool = True,
        norm: str = "l2",
        mc_samples: int = 10,
    ) -> xr.Dataset:
        """Compute local statistics from counterfactual sample representations.

        Local statistics are reductions over either the local counterfactual latent representations
        or the resulting local sample distance matrices. For a large number of cells and/or
        samples, this method can avoid scalability issues by grouping over cell-level covariates.

        Parameters
        ----------
        reductions
            List of reductions to compute over local counterfactual sample representations.
        adata
            AnnData object to use.
        indices
            Indices of cells to use.
        batch_size
            Batch size to use for computing the local statistics.
        use_vmap
            Whether to use vmap to compute the local statistics.
        norm
            Norm to use for computing the distances.
        mc_samples
            Number of Monte Carlo samples to use for computing the local statistics. Only applies
            if using sampled representations.
        """
        from functools import partial

        from scvi.external.mrvi._utils import _parse_local_statistics_requirements

        if not reductions or len(reductions) == 0:
            raise ValueError("At least one reduction must be provided.")

        adata = self.adata if adata is None else adata
        self._check_if_trained(warn=False)
        # Hack to ensure new AnnDatas have indices.
        adata.obs["_indices"] = np.arange(adata.n_obs).astype(int)

        adata = self._validate_anndata(adata)
        scdl = self._make_data_loader(
            adata=adata, indices=indices, batch_size=batch_size, iter_ndarray=True
        )
        n_sample = self.summary_stats.n_sample

        reqs = _parse_local_statistics_requirements(reductions)

        vars_in = {"params": self.module.params, **self.module.state}

        @partial(jax.jit, static_argnames=["use_mean", "mc_samples"])
        def mapped_inference_fn(
            stacked_rngs: dict[str, jax.random.KeyArray],
            x: jax.typing.ArrayLike,
            sample_index: jax.typing.ArrayLike,
            cf_sample: jax.typing.ArrayLike,
            use_mean: bool,
            mc_samples: int | None = None,
        ):
            # TODO: use `self.module.get_jit_inference_fn` when it supports traced values.
            def inference_fn(
                rngs,
                cf_sample,
            ):
                return self.module.apply(
                    vars_in,
                    rngs=rngs,
                    method=self.module.inference,
                    x=x,
                    sample_index=sample_index,
                    cf_sample=cf_sample,
                    use_mean=use_mean,
                    mc_samples=mc_samples,
                )["z"]

            if use_vmap:
                return jax.vmap(inference_fn, in_axes=(0, 0), out_axes=-2)(
                    stacked_rngs,
                    cf_sample,
                )
            else:

                def per_sample_inference_fn(pair):
                    rngs, cf_sample = pair
                    return inference_fn(rngs, cf_sample)

                return jax.lax.transpose(
                    jax.lax.map(per_sample_inference_fn, (stacked_rngs, cf_sample)),
                    (1, 0, 2),
                )

        ungrouped_data_arrs = {}
        grouped_data_arrs = {}
        for ur in reqs.ungrouped_reductions:
            ungrouped_data_arrs[ur.name] = []
        for gr in reqs.grouped_reductions:
            grouped_data_arrs[gr.name] = {}  # Will map group category to running group sum.

        for array_dict in tqdm(scdl):
            indices = array_dict[REGISTRY_KEYS.INDICES_KEY].astype(int).flatten()
            n_cells = array_dict[REGISTRY_KEYS.X_KEY].shape[0]
            cf_sample = np.broadcast_to(np.arange(n_sample)[:, None, None], (n_sample, n_cells, 1))
            inf_inputs = self.module._get_inference_input(
                array_dict,
            )
            stacked_rngs = self._generate_stacked_rngs(cf_sample.shape[0])

            # OK to use stacked rngs here since there is no stochasticity for mean rep.
            if reqs.needs_mean_representations:
                mean_zs_ = mapped_inference_fn(
                    stacked_rngs=stacked_rngs,
                    x=jnp.array(inf_inputs["x"]),
                    sample_index=jnp.array(inf_inputs["sample_index"]),
                    cf_sample=jnp.array(cf_sample),
                    use_mean=True,
                )
                mean_zs = xr.DataArray(
                    mean_zs_,
                    dims=["cell_name", "sample", "latent_dim"],
                    coords={
                        "cell_name": self.adata.obs_names[indices].values,
                        "sample": self.sample_order,
                    },
                    name="sample_representations",
                )
            if reqs.needs_sampled_representations:
                sampled_zs_ = mapped_inference_fn(
                    stacked_rngs=stacked_rngs,
                    x=jnp.array(inf_inputs["x"]),
                    sample_index=jnp.array(inf_inputs["sample_index"]),
                    cf_sample=jnp.array(cf_sample),
                    use_mean=False,
                    mc_samples=mc_samples,
                )  # (n_mc_samples, n_cells, n_samples, n_latent)
                sampled_zs_ = sampled_zs_.transpose((1, 0, 2, 3))
                sampled_zs = xr.DataArray(
                    sampled_zs_,
                    dims=["cell_name", "mc_sample", "sample", "latent_dim"],
                    coords={
                        "cell_name": self.adata.obs_names[indices].values,
                        "sample": self.sample_order,
                    },
                    name="sample_representations",
                )

            if reqs.needs_mean_distances:
                mean_dists = self._compute_distances_from_representations(
                    mean_zs_, indices, norm=norm
                )

            if reqs.needs_sampled_distances or reqs.needs_normalized_distances:
                sampled_dists = self._compute_distances_from_representations(
                    sampled_zs_, indices, norm=norm
                )

                if reqs.needs_normalized_distances:
                    if norm != "l2":
                        raise ValueError(
                            f"Norm must be 'l2' when using normalized distances. Got {norm}."
                        )
                    (
                        normalization_means,
                        normalization_vars,
                    ) = self._compute_local_baseline_dists(
                        array_dict, mc_samples=mc_samples
                    )  # both are shape (n_cells,)
                    normalization_means = normalization_means.reshape(-1, 1, 1, 1)
                    normalization_vars = normalization_vars.reshape(-1, 1, 1, 1)
                    normalized_dists = (
                        (sampled_dists - normalization_means) / (normalization_vars**0.5)
                    ).mean(dim="mc_sample")  # (n_cells, n_samples, n_samples)

            # Compute each reduction
            for r in reductions:
                if r.input == "mean_representations":
                    inputs = mean_zs
                elif r.input == "sampled_representations":
                    inputs = sampled_zs
                elif r.input == "mean_distances":
                    inputs = mean_dists
                elif r.input == "sampled_distances":
                    inputs = sampled_dists
                elif r.input == "normalized_distances":
                    inputs = normalized_dists
                else:
                    raise ValueError(f"Unknown reduction input: {r.input}")

                outputs = r.fn(inputs)

                if r.group_by is not None:
                    group_by = self.adata.obs[r.group_by].iloc[indices]
                    group_by_cats = group_by.unique()
                    for cat in group_by_cats:
                        cat_summed_outputs = outputs.sel(
                            cell_name=self.adata.obs_names[indices][group_by == cat].values
                        ).sum(dim="cell_name")
                        cat_summed_outputs = cat_summed_outputs.assign_coords(
                            {f"{r.group_by}_name": cat}
                        )
                        if cat not in grouped_data_arrs[r.name]:
                            grouped_data_arrs[r.name][cat] = cat_summed_outputs
                        else:
                            grouped_data_arrs[r.name][cat] += cat_summed_outputs
                else:
                    ungrouped_data_arrs[r.name].append(outputs)

        # Combine all outputs.
        final_data_arrs = {}
        for ur_name, ur_outputs in ungrouped_data_arrs.items():
            final_data_arrs[ur_name] = xr.concat(ur_outputs, dim="cell_name")

        for gr in reqs.grouped_reductions:
            group_by = adata.obs[gr.group_by]
            group_by_counts = group_by.value_counts()
            averaged_grouped_data_arrs = []
            for cat, count in group_by_counts.items():
                averaged_grouped_data_arrs.append(grouped_data_arrs[gr.name][cat] / count)
            final_data_arr = xr.concat(averaged_grouped_data_arrs, dim=f"{gr.group_by}_name")
            final_data_arrs[gr.name] = final_data_arr

        return xr.Dataset(data_vars=final_data_arrs)

    def _compute_local_baseline_dists(
        self, batch: dict, mc_samples: int = 250
    ) -> tuple[npt.NDArray, npt.NDArray]:
        """
        Approximate the distributions used as baselines for normalizing the local sample distances.

        Approximates the means and variances of the Euclidean distance between two samples of
        the z latent representation for the original sample for each cell in ``adata``.

        Reference: https://www.overleaf.com/read/mhdxcrknzxpm.

        Parameters
        ----------
        batch
            Batch of data to compute the local sample representation for.
        mc_samples
            Number of Monte Carlo samples to use for computing the local baseline distributions.
        """
        mc_samples_per_cell = (
            mc_samples * 2
        )  # need double for pairs of samples to compute distance between
        jit_inference_fn = self.module.get_jit_inference_fn(
            inference_kwargs={"use_mean": False, "mc_samples": mc_samples_per_cell}
        )

        outputs = jit_inference_fn(self.module.rngs, batch)

        # figure out how to compute dists here
        z = outputs["z"]
        first_half_z, second_half_z = z[:mc_samples], z[mc_samples:]
        l2_dists = jnp.sqrt(jnp.sum((first_half_z - second_half_z) ** 2, axis=2)).T

        return np.array(jnp.mean(l2_dists, axis=1)), np.array(jnp.var(l2_dists, axis=1))

    def _compute_distances_from_representations(
        self,
        reps: jax.typing.ArrayLike,
        indices: jax.typing.ArrayLike,
        norm: Literal["l2", "l1", "linf"] = "l2",
    ) -> xr.DataArray:
        if norm not in ("l2", "l1", "linf"):
            raise ValueError(f"`norm` {norm} not supported")

        @jax.jit
        def _compute_distance(rep: jax.typing.ArrayLike):
            delta_mat = jnp.expand_dims(rep, 0) - jnp.expand_dims(rep, 1)
            if norm == "l2":
                res = delta_mat**2
                res = jnp.sqrt(res.sum(-1))
            elif norm == "l1":
                res = jnp.abs(delta_mat).sum(-1)
            elif norm == "linf":
                res = jnp.abs(delta_mat).max(-1)
            return res

        if reps.ndim == 3:
            dists = jax.vmap(_compute_distance)(reps)
            return xr.DataArray(
                dists,
                dims=["cell_name", "sample_x", "sample_y"],
                coords={
                    "cell_name": self.adata.obs_names[indices].values,
                    "sample_x": self.sample_order,
                    "sample_y": self.sample_order,
                },
                name="sample_distances",
            )
        else:
            # Case with sampled representations
            dists = jax.vmap(jax.vmap(_compute_distance))(reps)
            return xr.DataArray(
                dists,
                dims=["cell_name", "mc_sample", "sample_x", "sample_y"],
                coords={
                    "cell_name": self.adata.obs_names[indices].values,
                    "mc_sample": np.arange(reps.shape[1]),
                    "sample_x": self.sample_order,
                    "sample_y": self.sample_order,
                },
                name="sample_distances",
            )

    def get_local_sample_representation(
        self,
        adata: AnnData | None = None,
        indices: npt.ArrayLike | None = None,
        batch_size: int = 256,
        use_mean: bool = True,
        use_vmap: bool = True,
    ) -> xr.DataArray:
        """Compute the local sample representation of the cells in the ``adata`` object.

        For each cell, it returns a matrix of size ``(n_sample, n_features)``.

        Parameters
        ----------
        adata
            AnnData object to use for computing the local sample representation.
        batch_size
            Batch size to use for computing the local sample representation.
        use_mean
            Whether to use the mean of the latent representation as the local sample
            representation.
        use_vmap
            Whether to use vmap for computing the local sample representation.
            Disabling vmap can be useful if running out of memory on a GPU.
        """
        reductions = [
            MRVIReduction(
                name="sample_representations",
                input="mean_representations" if use_mean else "sampled_representations",
                fn=lambda x: x,
                group_by=None,
            )
        ]
        return self.compute_local_statistics(
            reductions,
            adata=adata,
            indices=indices,
            batch_size=batch_size,
            use_vmap=use_vmap,
        ).sample_representations

    def get_local_sample_distances(
        self,
        adata: AnnData | None = None,
        batch_size: int = 256,
        use_mean: bool = True,
        normalize_distances: bool = False,
        use_vmap: bool = True,
        groupby: list[str] | str | None = None,
        keep_cell: bool = True,
        norm: str = "l2",
        mc_samples: int = 10,
    ) -> xr.Dataset:
        """Compute local sample distances.

        Computes cell-specific distances between samples, of size ``(n_sample, n_sample)``,
        stored as a Dataset, with variable name ``"cell"``, of size
        ``(n_cell, n_sample, n_sample)``. If in addition, ``groupby`` is provided, distances are
        also aggregated by group. In this case, the group-specific distances via group name key.

        Parameters
        ----------
        adata
            AnnData object to use for computing the local sample representation.
        batch_size
            Batch size to use for computing the local sample representation.
        use_mean
            Whether to use the mean of the latent representation as the local sample
            representation.
        normalize_distances
            Whether to normalize the local sample distances. Normalizes by the standard deviation
            of the original intra-sample distances. Only works with ``use_mean=False``.
        use_vmap
            Whether to use vmap for computing the local sample representation. Disabling vmap can
            be useful if running out of memory on a GPU.
        groupby
            List of categorical keys or single key of the anndata that is used to group the cells.
        keep_cell
            Whether to keep the original cell sample-sample distance matrices.
        norm
            Norm to use for computing the local sample distances.
        mc_samples
            Number of Monte Carlo samples to use for computing the local sample distances. Only
            relevant if ``use_mean=False``.
        """
        input = "mean_distances" if use_mean else "sampled_distances"
        if normalize_distances:
            if use_mean:
                warnings.warn(
                    "Normalizing distances uses sampled distances. Ignoring ``use_mean``.",
                    UserWarning,
                    stacklevel=2,
                )
            input = "normalized_distances"
        if groupby and not isinstance(groupby, list):
            groupby = [groupby]

        reductions = []
        if not keep_cell and not groupby:
            raise ValueError("Undefined computation because not keep_cell and no groupby.")
        if keep_cell:
            reductions.append(
                MRVIReduction(
                    name="cell",
                    input=input,
                    fn=lambda x: x,
                )
            )
        if groupby:
            for groupby_key in groupby:
                reductions.append(
                    MRVIReduction(
                        name=groupby_key,
                        input=input,
                        group_by=groupby_key,
                    )
                )
        return self.compute_local_statistics(
            reductions,
            adata=adata,
            batch_size=batch_size,
            use_vmap=use_vmap,
            norm=norm,
            mc_samples=mc_samples,
        )

    def get_aggregated_posterior(
        self,
        adata: AnnData | None = None,
        sample: str | int | None = None,
        indices: npt.ArrayLike | None = None,
        batch_size: int = 256,
    ) -> Distribution:
        """Computes the aggregated posterior over the ``u`` latent representations.

        For the specified samples, it computes the aggregated posterior over the ``u`` latent
        representations. Returns a NumPyro MixtureSameFamily distribution.

        Parameters
        ----------
        adata
            AnnData object to use. Defaults to the AnnData object used to initialize the model.
        sample
            Name or index of the sample to filter on. If ``None``, uses all cells.
        indices
            Indices of cells to use.
        batch_size
            Batch size to use for computing the latent representation.

        Returns
        -------
        A mixture distribution of the aggregated posterior.
        """
        from numpyro.distributions import Categorical, MixtureSameFamily, Normal

        self._check_if_trained(warn=False)
        adata = self._validate_anndata(adata)
        if indices is None:
            indices = np.arange(self.adata.n_obs)
        if sample is not None:
            indices = np.intersect1d(
                np.array(indices), np.where(adata.obs[self.sample_key] == sample)[0]
            )
        scdl = self._make_data_loader(
            adata=adata, indices=indices, batch_size=batch_size, iter_ndarray=True
        )

        qu_locs = []
        qu_scales = []
        jit_inference_fn = self.module.get_jit_inference_fn(inference_kwargs={"use_mean": True})
        for array_dict in scdl:
            outputs = jit_inference_fn(self.module.rngs, array_dict)

            qu_locs.append(outputs["qu"].loc)
            qu_scales.append(outputs["qu"].scale)

        qu_loc = jnp.concatenate(qu_locs, axis=0).T
        qu_scale = jnp.concatenate(qu_scales, axis=0).T
        return MixtureSameFamily(
            Categorical(probs=jnp.ones(qu_loc.shape[1]) / qu_loc.shape[1]),
            Normal(qu_loc, qu_scale),
        )

    def differential_abundance(
        self,
        adata: AnnData | None = None,
        sample_cov_keys: list[str] | None = None,
        sample_subset: list[str] | None = None,
        compute_log_enrichment: bool = False,
        batch_size: int = 128,
    ) -> xr.Dataset:
        """Compute the differential abundance between samples.

        Computes the logarithm of the ratio of the probabilities of each sample conditioned on the
        estimated aggregate posterior distribution of each cell.

        Parameters
        ----------
        adata
            The data object to compute the differential abundance for.
            If not given, the data object stored in the model is used.
        sample_cov_keys
            Keys for covariates (batch, etc.) that should also be taken into account
            when computing the differential abundance. At the moment, only discrete covariates are
            supported.
        sample_subset
            Only compute differential abundance for these sample labels.
        compute_log_enrichment
            Whether to compute the log enrichment scores for each covariate value.
        batch_size
            Minibatch size for computing the differential abundance.

        Returns
        -------
        A dataset with data variables:

        * ``"log_probs"``: Array of shape ``(n_cells, n_samples)`` containing the log probabilities
            for each cell across samples.
        * ``"{cov_key}_log_probs"``: For each key in ``sample_cov_keys``, an array of shape
            ``(n_cells, _cov_values)`` containing the log probabilities for each cell across
            covariate values.
        """
        from pandas import DataFrame
        from scipy.special import logsumexp

        adata = self._validate_anndata(adata)

        if sample_cov_keys is not None:
            for key in sample_cov_keys:
                n_cov_values = len(adata.obs[key].unique())
                n_samples = len(adata.obs[self.sample_key].unique())
                if n_cov_values > n_samples / 2:
                    warnings.warn(
                        f"The covariate '{key}' does not seem to refer to a discrete key. "
                        f"It has {n_cov_values} unique values, which exceeds one half of the "
                        f"total samples ({n_samples}).",
                        UserWarning,
                        stacklevel=2,
                    )

        us = self.get_latent_representation(
            adata, use_mean=True, give_z=False, batch_size=batch_size
        )

        log_probs = []
        unique_samples = adata.obs[self.sample_key].unique()
        for sample_name in tqdm(unique_samples):
            ap = self.get_aggregated_posterior(
                adata=adata, sample=sample_name, batch_size=batch_size
            )
            n_splits = max(adata.n_obs // batch_size, 1)
            log_probs_ = []
            for u_rep in np.array_split(us, n_splits):
                log_probs_.append(jax.device_get(ap.log_prob(u_rep).sum(-1, keepdims=True)))

            log_probs.append(np.concatenate(log_probs_, axis=0))  # (n_cells, 1)

        log_probs = np.concatenate(log_probs, 1)

        coords = {
            "cell_name": adata.obs_names.to_numpy(),
            "sample": unique_samples,
        }
        data_vars = {
            "log_probs": (["cell_name", "sample"], log_probs),
        }
        log_probs_arr = xr.Dataset(data_vars, coords=coords)

        if sample_cov_keys is None or len(sample_cov_keys) == 0:
            return log_probs_arr

        sample_cov_log_probs_map = {}
        sample_cov_log_enrichs_map = {}
        for sample_cov_key in sample_cov_keys:
            sample_cov_unique_values = self.sample_info[sample_cov_key].unique()
            per_val_log_probs = {}
            per_val_log_enrichs = {}
            for sample_cov_value in sample_cov_unique_values:
                cov_samples = (
                    self.sample_info[self.sample_info[sample_cov_key] == sample_cov_value]
                )[self.sample_key].to_numpy()
                if sample_subset is not None:
                    cov_samples = np.intersect1d(cov_samples, np.array(sample_subset))
                if len(cov_samples) == 0:
                    continue

                sel_log_probs = log_probs_arr.log_probs.loc[{"sample": cov_samples}]
                val_log_probs = logsumexp(sel_log_probs, axis=1) - np.log(sel_log_probs.shape[1])
                per_val_log_probs[sample_cov_value] = val_log_probs

                if compute_log_enrichment:
                    rest_samples = np.setdiff1d(unique_samples, cov_samples)
                    if len(rest_samples) == 0:
                        warnings.warn(
                            f"All samples have {sample_cov_key}={sample_cov_value}. Skipping log "
                            "enrichment computation.",
                            UserWarning,
                            stacklevel=2,
                        )
                        continue
                    rest_log_probs = log_probs_arr.log_probs.loc[{"sample": rest_samples}]
                    rest_val_log_probs = logsumexp(rest_log_probs, axis=1) - np.log(
                        rest_log_probs.shape[1]
                    )
                    enrichment_scores = val_log_probs - rest_val_log_probs
                    per_val_log_enrichs[sample_cov_value] = enrichment_scores
            sample_cov_log_probs_map[sample_cov_key] = DataFrame.from_dict(per_val_log_probs)
            if compute_log_enrichment and len(per_val_log_enrichs) > 0:
                sample_cov_log_enrichs_map[sample_cov_key] = DataFrame.from_dict(
                    per_val_log_enrichs
                )

        coords = {
            "cell_name": adata.obs_names.to_numpy(),
            "sample": unique_samples,
            **{
                sample_cov_key: sample_cov_log_probs.columns
                for sample_cov_key, sample_cov_log_probs in sample_cov_log_probs_map.items()
            },
        }
        data_vars = {
            "log_probs": (["cell_name", "sample"], log_probs),
            **{
                f"{sample_cov_key}_log_probs": (
                    ["cell_name", sample_cov_key],
                    sample_cov_log_probs.values,
                )
                for sample_cov_key, sample_cov_log_probs in sample_cov_log_probs_map.items()
            },
        }
        if compute_log_enrichment:
            data_vars.update(
                {
                    f"{sample_key}_log_enrichs": (
                        ["cell_name", sample_key],
                        sample_log_enrichs.values,
                    )
                    for sample_key, sample_log_enrichs in sample_cov_log_enrichs_map.items()
                }
            )
        return xr.Dataset(data_vars, coords=coords)

    def get_outlier_cell_sample_pairs(
        self,
        adata: AnnData | None = None,
        subsample_size: int = 5_000,
        quantile_threshold: float = 0.05,
        admissibility_threshold: float = 0.0,
        batch_size: int = 256,
    ) -> xr.Dataset:
        """Compute admissibility scores for cell-sample pairs.

        This function computes the posterior distribution for u for each cell. Then, for every
        cell, it computes the log-probability of the cell under the posterior of each cell
        each sample and takes the maximum value for a given sample as a measure of admissibility
        for that sample. Additionally, it computes a threshold that determines if
        a cell-sample pair is admissible based on the within-sample admissibility scores.

        Parameters
        ----------
        adata
            AnnData object containing the cells for which to compute the outlier cell-sample pairs.
        subsample_size
            Number of cells to use from each sample to approximate the posterior. If None, uses all
            of the available cells.
        quantile_threshold
            Quantile of the within-sample log probabilities to use as a baseline for admissibility.
        admissibility_threshold
            Threshold for admissibility. Cell-sample pairs with admissibility below this threshold
            are considered outliers.
        batch_size
            Size of the batch to use for computing outlier cell-sample pairs.
        """
        adata = self._validate_anndata(adata)
        us = self.get_latent_representation(adata, use_mean=True, give_z=False)
        adata.obsm["U"] = us

        log_probs = []
        threshs = []
        unique_samples = adata.obs[self.sample_key].unique()
        for sample_name in tqdm(unique_samples):
            sample_idxs = np.where(adata.obs[self.sample_key] == sample_name)[0]
            if subsample_size is not None and sample_idxs.shape[0] > subsample_size:
                sample_idxs = np.random.choice(sample_idxs, size=subsample_size, replace=False)
            adata_s = adata[sample_idxs]

            ap = self.get_aggregated_posterior(adata=adata, indices=sample_idxs)
            in_max_comp_log_probs = ap.component_distribution.log_prob(
                np.expand_dims(adata_s.obsm["U"], ap.mixture_dim)
            ).sum(axis=1)
            log_probs_s = rowwise_max_excluding_diagonal(in_max_comp_log_probs)

            log_probs_ = []
            n_splits = adata.n_obs // batch_size
            for u_rep in np.array_split(adata.obsm["U"], n_splits):
                log_probs_.append(
                    jax.device_get(
                        ap.component_distribution.log_prob(
                            np.expand_dims(u_rep, ap.mixture_dim)
                        )  # (n_cells_batch, n_cells_ap, n_latent_dim)
                        .sum(axis=1)  # (n_cells_batch, n_latent_dim)
                        .max(axis=1, keepdims=True)  # (n_cells_batch, 1)
                    )
                )

            log_probs_ = np.concatenate(log_probs_, axis=0)  # (n_cells, 1)

            threshs.append(np.array(log_probs_s))
            log_probs.append(np.array(log_probs_))

        threshs_all = np.concatenate(threshs)
        global_thresh = np.quantile(threshs_all, q=quantile_threshold)
        threshs = np.array(len(log_probs) * [global_thresh])

        log_probs = np.concatenate(log_probs, 1)
        log_ratios = log_probs - threshs

        coords = {
            "cell_name": adata.obs_names.to_numpy(),
            "sample": unique_samples,
        }
        data_vars = {
            "log_probs": (["cell_name", "sample"], log_probs),
            "log_ratios": (
                ["cell_name", "sample"],
                log_ratios,
            ),
            "is_admissible": (
                ["cell_name", "sample"],
                log_ratios > admissibility_threshold,
            ),
        }
        return xr.Dataset(data_vars, coords=coords)

    def differential_expression(
        self,
        adata: AnnData | None = None,
        sample_cov_keys: list[str] | None = None,
        sample_subset: list[str] | None = None,
        batch_size: int = 128,
        use_vmap: bool = True,
        normalize_design_matrix: bool = True,
        add_batch_specific_offsets: bool = False,
        mc_samples: int = 100,
        store_lfc: bool = False,
        store_lfc_metadata_subset: list[str] | None = None,
        store_baseline: bool = False,
        eps_lfc: float = 1e-4,
        filter_inadmissible_samples: bool = False,
        lambd: float = 0.0,
        delta: float | None = 0.3,
        **filter_samples_kwargs,
    ) -> xr.Dataset:
        """Compute cell-specific multivariate differential expression.

        For every cell, we first compute all counterfactual cell-state shifts, defined as
        ``e_d = z_d - u``, where ``z_d`` is the latent representation of the cell for sample ``d``
        and ``u`` is the sample-unaware latent representation. Then, we fit a linear model in each
        cell of the form: ``e_d = X_d * beta_d + iid gaussian noise``.

        Parameters
        ----------
        sample_cov_keys
            List of sample covariates to consider for the multivariate analysis.
            These keys should be present in ``adata.obs``.
        adata
            AnnData object to use for computing the local sample representation.
            If ``None``, the analysis is performed on all cells in the dataset.
        sample_subset
            Optional list of samples to consider for the multivariate analysis.
            If ``None``, all samples are considered.
        batch_size
            Batch size to use for computing the local sample representation.
        use_vmap
            Whether to use vmap for computing the local sample representation.
        normalize_design_matrix
            Whether to normalize the design matrix.
        add_batch_specific_offsets
            Whether to offset the design matrix by adding batch-specific offsets to the design
            matrix. Setting this option to True is recommended when considering multi-site
            datasets.
        mc_samples
            How many MC samples should be taken for computing betas.
        store_lfc
            Whether to store the log-fold changes in the module.
            Storing log-fold changes is memory-intensive and may require to specify
            a smaller set of cells to analyze, e.g., by specifying ``adata``.
        store_lfc_metadata_subset
            Specifies a subset of metadata for which log-fold changes are computed.
            These keys must be a subset of ``sample_cov_keys``.
            Only applies when ``store_lfc=True``.
        store_baseline
            Whether to store the expression in the module if logfoldchanges are computed.
        eps_lfc
            Epsilon to add to the log-fold changes to avoid detecting genes with low expression.
        filter_inadmissible_samples
            Whether to filter out-of-distribution samples prior to performing the analysis.
        lambd
            Regularization parameter for the linear model.
        delta
            LFC threshold used to compute posterior DE probabilities.
            If None does not compute them to save memory consumption.
        filter_samples_kwargs
            Keyword arguments to pass to :meth:`~scvi.external.MRVI.get_outlier_cell_sample_pairs`.

        Returns
        -------
        A dataset containing the results of the differential expression analysis:

        * ``"beta"``: Coefficients for each covariate across cells and latent dimensions.
        * ``"effect_size"``: Effect sizes for each covariate across cells.
        * ``"pvalue"``: P-values for each covariate across cells.
        * ``"padj"``: Adjusted P-values for each covariate across cells using the
            Benjamini-Hochberg procedure.
        * ``"lfc"``: Log fold changes for each covariate across cells and genes, if ``store_lfc``
            is ``True``.
        * ``"lfc_std"``: Standard deviation of log fold changes, if ``store_lfc`` is ``True`` and
            ``delta`` is not ``None``.
        * ``"pde"``: Posterior DE probabilities, if ``store_lfc`` is ``True`` and ``delta`` is not
            ``None``.
        * ``"baseline_expression"``: Baseline expression levels for each covariate across cells and
            genes, if ``store_baseline`` is ``True``.
        * ``"n_samples"``: Number of admissible samples for each cell, if
            ``filter_inadmissible_samples`` is ``True``.
        """
        from functools import partial

        from scipy.stats import false_discovery_control

        if sample_cov_keys is None:
            # Hack: kept as kwarg to maintain order of arguments.
            raise ValueError("Must assign `sample_cov_keys`")
        adata = self.adata if adata is None else adata
        self._check_if_trained(warn=False)
        # Hack to ensure new AnnDatas have indices and indices have correct dimensions.
        if adata is not None:
            adata.obs["_indices"] = np.arange(adata.n_obs).astype(int)

        adata = self._validate_anndata(adata)
        scdl = self._make_data_loader(
            adata=adata, indices=None, batch_size=batch_size, iter_ndarray=True
        )
        n_sample = self.summary_stats.n_sample
        vars_in = {"params": self.module.params, **self.module.state}

        sample_mask = (
            np.isin(self.sample_order, sample_subset)
            if sample_subset is not None
            else np.ones(n_sample, dtype=bool)
        )
        sample_mask = np.array(sample_mask)
        sample_order = self.sample_order[sample_mask]
        n_samples_kept = sample_mask.sum()

        if filter_inadmissible_samples:
            admissible_samples = self.get_outlier_cell_sample_pairs(
                adata=adata, **filter_samples_kwargs
            )["is_admissible"].loc[{"sample": sample_order}]
            assert (admissible_samples.sample == sample_order).all()
            admissible_samples = admissible_samples.values
        else:
            admissible_samples = np.ones((adata.n_obs, n_samples_kept), dtype=bool)
        n_admissible_samples = admissible_samples.sum(1)

        (
            Xmat,
            Xmat_names,
            covariates_require_lfc,
            offset_indices,
        ) = self._construct_design_matrix(
            sample_cov_keys=sample_cov_keys,
            sample_mask=sample_mask,
            normalize_design_matrix=normalize_design_matrix,
            add_batch_specific_offsets=add_batch_specific_offsets,
            store_lfc=store_lfc,
            store_lfc_metadata_subset=store_lfc_metadata_subset,
        )
        add_batch_specific_offsets = offset_indices is not None
        n_covariates = Xmat.shape[1]

        @partial(jax.jit, backend="cpu")
        def process_design_matrix(
            admissible_samples_dmat: jax.typing.ArrayLike,
            Xmat: jax.typing.ArrayLike,
        ) -> tuple[jax.Array, jax.Array]:
            xtmx = jnp.einsum("ak,nkl,lm->nam", Xmat.T, admissible_samples_dmat, Xmat)
            xtmx += lambd * jnp.eye(n_covariates)

            prefactor = jnp.real(jax.vmap(jax.scipy.linalg.sqrtm)(xtmx))
            inv_ = jax.vmap(jnp.linalg.pinv)(xtmx)
            Amat = jnp.einsum("nab,bc,ncd->nad", inv_, Xmat.T, admissible_samples_dmat)
            return Amat, prefactor

        @partial(jax.jit, static_argnames=["use_mean", "mc_samples"])
        def mapped_inference_fn(
            stacked_rngs: dict[str, jax.random.KeyArray],
            x: jax.typing.ArrayLike,
            sample_index: jax.typing.ArrayLike,
            cf_sample: jax.typing.ArrayLike,
            Amat: jax.typing.ArrayLike,
            prefactor: jax.typing.ArrayLike,
            n_samples_per_cell: int,
            admissible_samples_mat: jax.typing.ArrayLike,
            use_mean: bool,
            mc_samples: int,
            rngs_de=None,
        ):
            def inference_fn(
                rngs,
                cf_sample,
            ):
                return self.module.apply(
                    vars_in,
                    rngs=rngs,
                    method=self.module.inference,
                    x=x,
                    sample_index=sample_index,
                    cf_sample=cf_sample,
                    use_mean=use_mean,
                    mc_samples=mc_samples,
                )["eps"]

            if use_vmap:
                eps_ = jax.vmap(inference_fn, in_axes=(0, 0), out_axes=-2)(
                    stacked_rngs,
                    cf_sample,
                )
            else:

                def per_sample_inference_fn(pair):
                    rngs, cf_sample = pair
                    return inference_fn(rngs, cf_sample)

                # eps_ has shape (mc_samples, n_cells, n_samples, n_latent)
                eps_ = jax.lax.transpose(
                    jax.lax.map(per_sample_inference_fn, (stacked_rngs, cf_sample)),
                    (1, 2, 0, 3),
                )
            eps_std = eps_.std(axis=2, keepdims=True)
            eps_mean = eps_.mean(axis=2, keepdims=True)

            eps = (eps_ - eps_mean) / (1e-6 + eps_std)  # over samples
            # MLE for betas
            betas = jnp.einsum("nks,ansd->ankd", Amat, eps)

            # Statistical tests
            betas_norm = jnp.einsum("ankd,nkl->anld", betas, prefactor)
            ts = (betas_norm**2).mean(axis=0).sum(axis=-1)
            pvals = 1 - jnp.nan_to_num(
                jax.scipy.stats.chi2.cdf(ts, df=n_samples_per_cell[:, None]), nan=0.0
            )

            betas = betas * eps_std

            lfc_mean = None
            lfc_std = None
            pde = None
            if store_lfc:
                betas_ = betas.transpose((0, 2, 1, 3))
                eps_mean_ = eps_mean.transpose((0, 2, 1, 3))
                betas_covariates = betas_[:, covariates_require_lfc, :, :]

                def h_inference_fn(extra_eps, batch_index_cf, batch_offset_eps):
                    extra_eps += batch_offset_eps

                    return self.module.apply(
                        vars_in,
                        rngs=rngs_de,
                        method=self.module.compute_h_from_x_eps,
                        x=x,
                        extra_eps=extra_eps,
                        sample_index=sample_index,
                        batch_index=batch_index_cf,
                        cf_sample=None,
                        mc_samples=None,  # mc_samples also taken for eps. vmap over mc_samples
                    )

                batch_index_ = jnp.arange(self.summary_stats.n_batch)[:, None]
                batch_index_ = jnp.repeat(batch_index_, repeats=n_cells, axis=1)[
                    ..., None
                ]  # (n_batch, n_cells, 1)
                betas_null = jnp.zeros_like(betas_covariates)

                if add_batch_specific_offsets:
                    batch_weights = jnp.einsum(
                        "nd,db->nb", admissible_samples_mat, Xmat[:, offset_indices]
                    ).mean(0)
                    betas_offset_ = betas_[:, offset_indices, :, :] + eps_mean_
                else:
                    batch_weights = (1.0 / self.summary_stats.n_batch) * jnp.ones(
                        self.summary_stats.n_batch
                    )
                    mc_samples, _, n_cells_, n_latent = betas_covariates.shape
                    betas_offset_ = (
                        jnp.zeros((mc_samples, self.summary_stats.n_batch, n_cells_, n_latent))
                        + eps_mean_
                    )
                # batch_offset shape (mc_samples, n_batch, n_cells, n_latent)

                f_ = jax.vmap(
                    h_inference_fn, in_axes=(0, None, 0), out_axes=0
                )  # fn over MC samples
                f_ = jax.vmap(f_, in_axes=(1, None, None), out_axes=1)  # fn over covariates
                f_ = jax.vmap(f_, in_axes=(None, 0, 1), out_axes=0)  # fn over batches
                h_fn = jax.jit(f_)

                x_1 = h_fn(betas_covariates, batch_index_, betas_offset_)
                x_0 = h_fn(betas_null, batch_index_, betas_offset_)

                lfcs = jnp.log2(x_1 + eps_lfc) - jnp.log2(x_0 + eps_lfc)
                lfc_mean = jnp.average(lfcs.mean(1), weights=batch_weights, axis=0)
                if delta is not None:
                    lfc_std = jnp.sqrt(jnp.average(lfcs.var(1), weights=batch_weights, axis=0))
                    pde = (jnp.abs(lfcs) >= delta).mean(1).mean(0)

            if store_baseline:
                baseline_expression = x_1.mean(1)
            else:
                baseline_expression = None
            return {
                "beta": betas.mean(0),
                "effect_size": ts,
                "pvalue": pvals,
                "lfc_mean": lfc_mean,
                "lfc_std": lfc_std,
                "pde": pde,
                "baseline_expression": baseline_expression,
            }

        beta = []
        effect_size = []
        pvalue = []
        lfc = []
        lfc_std = []
        pde = []
        baseline_expression = []
        for array_dict in tqdm(scdl):
            indices = array_dict[REGISTRY_KEYS.INDICES_KEY].astype(int).flatten()
            n_cells = array_dict[REGISTRY_KEYS.X_KEY].shape[0]
            cf_sample = np.broadcast_to(
                (np.where(sample_mask)[0])[:, None, None], (n_samples_kept, n_cells, 1)
            )
            inf_inputs = self.module._get_inference_input(
                array_dict,
            )
            stacked_rngs = self._generate_stacked_rngs(cf_sample.shape[0])

            rngs_de = self.module.rngs if store_lfc else None
            admissible_samples_mat = jnp.array(admissible_samples[indices])  # (n_cells, n_samples)
            n_samples_per_cell = admissible_samples_mat.sum(axis=1)
            admissible_samples_dmat = jax.vmap(jnp.diag)(admissible_samples_mat).astype(
                float
            )  # (n_cells, n_samples, n_samples)
            # element nij is 1 if sample i is admissible and i=j for cell n
            Amat, prefactor = process_design_matrix(admissible_samples_dmat, Xmat)
            Amat = jax.device_put(Amat, self.device)
            prefactor = jax.device_put(prefactor, self.device)

            res = mapped_inference_fn(
                stacked_rngs=stacked_rngs,
                x=jnp.array(inf_inputs["x"]),
                sample_index=jnp.array(inf_inputs["sample_index"]),
                cf_sample=jnp.array(cf_sample),
                Amat=Amat,
                prefactor=prefactor,
                n_samples_per_cell=n_samples_per_cell,
                admissible_samples_mat=admissible_samples_mat,
                use_mean=False,
                rngs_de=rngs_de,
                mc_samples=mc_samples,
            )
            beta.append(np.array(res["beta"]))
            effect_size.append(np.array(res["effect_size"]))
            pvalue.append(np.array(res["pvalue"]))
            if store_lfc:
                lfc.append(np.array(res["lfc_mean"]))
                if delta is not None:
                    lfc_std.append(np.array(res["lfc_std"]))
                    pde.append(np.array(res["pde"]))
            if store_baseline:
                baseline_expression.append(np.array(res["baseline_expression"]))
        beta = np.concatenate(beta, axis=0)
        effect_size = np.concatenate(effect_size, axis=0)
        pvalue = np.concatenate(pvalue, axis=0)
        pvalue_shape = pvalue.shape
        padj = false_discovery_control(pvalue.flatten(), method="bh").reshape(pvalue_shape)

        coords = {
            "cell_name": (("cell_name"), adata.obs_names),
            "covariate": (("covariate"), Xmat_names),
            "latent_dim": (("latent_dim"), np.arange(beta.shape[2])),
            "gene": (("gene"), adata.var_names),
        }
        data_vars = {
            "beta": (
                ["cell_name", "covariate", "latent_dim"],
                beta,
            ),
            "effect_size": (
                ["cell_name", "covariate"],
                effect_size,
            ),
            "pvalue": (
                ["cell_name", "covariate"],
                pvalue,
            ),
            "padj": (
                ["cell_name", "covariate"],
                padj,
            ),
        }
        if filter_inadmissible_samples:
            data_vars["n_samples"] = (
                ["cell_name"],
                n_admissible_samples,
            )
        if store_lfc:
            if store_lfc_metadata_subset is None and not add_batch_specific_offsets:
                coords_lfc = ["covariate", "cell_name", "gene"]
            else:
                coords_lfc = ["covariate_sub", "cell_name", "gene"]
                coords["covariate_sub"] = (
                    ("covariate_sub"),
                    Xmat_names[covariates_require_lfc],
                )
            lfc = np.concatenate(lfc, axis=1)
            data_vars["lfc"] = (coords_lfc, lfc)
            if delta is not None:
                lfc_std = np.concatenate(lfc_std, axis=1)
                pde = np.concatenate(pde, axis=1)
                data_vars["lfc_std"] = (coords_lfc, lfc_std)
                data_vars["pde"] = (coords_lfc, pde)

        if store_baseline:
            baseline_expression = np.concatenate(baseline_expression, axis=1)
            data_vars["baseline_expression"] = (
                ["covariate", "cell_name", "gene"],
                baseline_expression,
            )
        return xr.Dataset(data_vars, coords=coords)

    def _construct_design_matrix(
        self,
        sample_cov_keys: list[str],
        sample_mask: npt.ArrayLike,
        normalize_design_matrix: bool,
        add_batch_specific_offsets: bool,
        store_lfc: bool,
        store_lfc_metadata_subset: list[str] | None = None,
    ) -> tuple[jax.Array, npt.NDArray, jax.Array, jax.Array | None]:
        """Construct a design matrix of samples and covariates.

        Starting from a list of sample covariate keys, construct a design matrix of samples and
        covariates. Categorical covariates are one-hot encoded.

        Parameters
        ----------
        sample_cov_keys
            List of sample metadata to use as covariates.
        sample_mask
            Mask of admissible samples. Must have the same length as the number of samples in the
            dataset.
        normalize_design_matrix
            Whether the design matrix should be 0-1 normalized. This is useful to ensure that the
            beta coefficients are comparable across covariates.
        add_batch_specific_offsets
            Whether the design matrix should be offset. If True, the matrix includes batch-specific
            offsets. This ensures that we can learn perturbation effects that do not depend on
            batch effects.

        Returns
        -------
        A tuple consisting of:

        1. The design matrix
        2. Names for each column in the design matrix
        3. A mask precising which coefficients from the design matrix require to compute LFCs.
        4. A mask precising which coefficients from the design matrix correspond to offsets.
        """
        from pandas import Series, get_dummies

        Xmat = []
        Xmat_names = []
        Xmat_dim_to_key = []
        sample_info = self.sample_info.iloc[sample_mask]
        for sample_cov_key in tqdm(sample_cov_keys):
            cov = sample_info[sample_cov_key]
            if (cov.dtype == str) or (cov.dtype == "category"):
                cov = cov.cat.remove_unused_categories()
                cov = get_dummies(cov, drop_first=True)
                cov_names = np.array([f"{sample_cov_key}_{col}" for col in cov.columns])
                cov = cov.values
            else:
                cov_names = np.array([sample_cov_key])
                cov = cov.values[:, None]
            n_covs = cov.shape[1]
            Xmat.append(cov)
            Xmat_names.append(cov_names)
            Xmat_dim_to_key.append([sample_cov_key] * n_covs)
        Xmat_names = np.concatenate(Xmat_names)
        Xmat = np.concatenate(Xmat, axis=1).astype(np.float32)
        Xmat_dim_to_key = np.concatenate(Xmat_dim_to_key)

        if normalize_design_matrix:
            Xmat = (Xmat - Xmat.min(axis=0)) / (1e-6 + Xmat.max(axis=0) - Xmat.min(axis=0))
        if add_batch_specific_offsets:
            cov = sample_info["_scvi_batch"]
            if cov.nunique() == self.summary_stats.n_batch:
                cov = np.eye(self.summary_stats.n_batch)[sample_info["_scvi_batch"].values]
                cov_names = ["offset_batch_" + str(i) for i in range(self.summary_stats.n_batch)]
                Xmat = np.concatenate([cov, Xmat], axis=1)
                Xmat_names = np.concatenate([np.array(cov_names), Xmat_names])
                Xmat_dim_to_key = np.concatenate([np.array(cov_names), Xmat_dim_to_key])

                # Retrieve indices of offset covariates in the right order
                offset_indices = (
                    Series(np.arange(len(Xmat_names)), index=Xmat_names).loc[cov_names].values
                )
                offset_indices = jnp.array(offset_indices)
            else:
                warnings.warn(
                    """
                        Number of batches in sample_info does not match number of batches in
                        summary_stats. `add_batch_specific_offsets=True` assumes that samples are
                        not shared across batches. Setting `add_batch_specific_offsets=False`...
                    """,
                    stacklevel=2,
                )
                offset_indices = None
        else:
            offset_indices = None

        Xmat = jnp.array(Xmat)
        if store_lfc:
            covariates_require_lfc = (
                np.isin(Xmat_dim_to_key, store_lfc_metadata_subset)
                if store_lfc_metadata_subset is not None
                else np.isin(Xmat_dim_to_key, sample_cov_keys)
            )
        else:
            covariates_require_lfc = np.zeros(len(Xmat_names), dtype=bool)
        covariates_require_lfc = jnp.array(covariates_require_lfc)

        return Xmat, Xmat_names, covariates_require_lfc, offset_indices



import os

import numpy as np

from scvi.data import synthetic_iid
from scvi.model import AmortizedLDA


def test_lda_model_single_step(n_topics: int = 5):
    adata = synthetic_iid()
    AmortizedLDA.setup_anndata(adata)
    mod1 = AmortizedLDA(adata, n_topics=n_topics, cell_topic_prior=1.5, topic_feature_prior=1.5)
    mod1.train(max_steps=1, max_epochs=10)
    assert len(mod1.history["elbo_train"]) == 1


def test_lda_model_single_step_with_external_indices(n_topics: int = 5):
    adata = synthetic_iid()
    AmortizedLDA.setup_anndata(adata)
    mod1 = AmortizedLDA(adata, n_topics=n_topics, cell_topic_prior=1.5, topic_feature_prior=1.5)
    # in this case we will make a stratified version of indexing
    from sklearn.model_selection import train_test_split

    train_ind, valid_ind = train_test_split(
        adata.obs.batch.index.astype(int), test_size=0.6, stratify=adata.obs.batch
    )
    test_ind, valid_ind = train_test_split(
        valid_ind, test_size=0.5, stratify=adata.obs.batch[valid_ind]
    )
    mod1.train(
        max_steps=1,
        max_epochs=10,
        datasplitter_kwargs={
            "external_indexing": [np.array(train_ind), np.array(valid_ind), np.array(test_ind)]
        },
    )
    assert len(mod1.history["elbo_train"]) == 1


def test_lda_model(n_topics: int = 5):
    adata = synthetic_iid()

    # Test with float and Sequence priors.
    AmortizedLDA.setup_anndata(adata)
    mod1 = AmortizedLDA(adata, n_topics=n_topics, cell_topic_prior=1.5, topic_feature_prior=1.5)
    mod1.train(
        max_epochs=1,
        batch_size=256,
        lr=0.01,
    )
    mod2 = AmortizedLDA(
        adata,
        n_topics=n_topics,
        cell_topic_prior=[1.5 for _ in range(n_topics)],
        topic_feature_prior=[1.5 for _ in range(adata.n_vars)],
    )
    mod2.train(
        max_epochs=1,
        batch_size=256,
        lr=0.01,
    )

    mod = AmortizedLDA(adata, n_topics=n_topics)
    mod.train(
        max_epochs=5,
        batch_size=256,
        lr=0.01,
    )
    adata_gbt = mod.get_feature_by_topic().to_numpy()
    assert np.allclose(adata_gbt.sum(axis=0), 1)
    adata_lda = mod.get_latent_representation(adata).to_numpy()
    assert adata_lda.shape == (adata.n_obs, n_topics)
    assert np.all((adata_lda <= 1) & (adata_lda >= 0))
    assert np.allclose(adata_lda.sum(axis=1), 1)
    mod.get_elbo()
    mod.get_perplexity()

    adata2 = synthetic_iid()
    AmortizedLDA.setup_anndata(adata2)
    adata2_lda = mod.get_latent_representation(adata2).to_numpy()
    assert adata2_lda.shape == (adata2.n_obs, n_topics)
    assert np.all((adata2_lda <= 1) & (adata2_lda >= 0))
    assert np.allclose(adata2_lda.sum(axis=1), 1)
    mod.get_elbo(adata2)
    mod.get_perplexity(adata2)


def test_lda_model_save_load(save_path: str, n_topics: int = 5):
    adata = synthetic_iid()
    AmortizedLDA.setup_anndata(adata)
    mod = AmortizedLDA(adata, n_topics=n_topics)
    mod.train(
        max_epochs=5,
        batch_size=256,
        lr=0.01,
    )
    hist_elbo = mod.history_["elbo_train"]

    feature_by_topic_1 = mod.get_feature_by_topic(n_samples=5000)
    latent_1 = mod.get_latent_representation(n_samples=6000)

    save_path = os.path.join(save_path, "tmp")
    mod.save(save_path, overwrite=True, save_anndata=True)
    mod = AmortizedLDA.load(save_path)

    np.testing.assert_array_equal(mod.history_["elbo_train"], hist_elbo)

    feature_by_topic_2 = mod.get_feature_by_topic(n_samples=5000)
    latent_2 = mod.get_latent_representation(n_samples=6000)
    np.testing.assert_almost_equal(
        feature_by_topic_1.to_numpy(), feature_by_topic_2.to_numpy(), decimal=2
    )
    np.testing.assert_almost_equal(latent_1.to_numpy(), latent_2.to_numpy(), decimal=2)



import numpy as np
from anndata import AnnData


def _rot(t, u=1):
    if u not in [-1, 1]:
        raise ValueError("u must be -1 or 1")
    return np.array([[np.cos(t), np.sin(t) * u], [-np.sin(t), np.cos(t) * u]])


def rotate_decipher_components(
    adata: AnnData,
    v_obsm_key: str,
    z_obsm_key: str,
    v1_obs_col: str | None = None,
    v1_order: list | None = None,
    v2_obs_col: str | None = None,
    v2_order: list | None = None,
    auto_flip_decipher_z: bool = True,
) -> tuple[np.array, np.array, np.array]:
    """
    Rotate and flip the decipher space v to maximize the correlation of each decipher component.

    The rotation is performed to maximize the correlation of each decipher component with the
    providedcolumns values from `adata.obs` (e.g. pseudo-time, cell state progression, etc.).

    Parameters
    ----------
    adata: sc.AnnData
        The annotated data matrix.
    v_obsm_key: str
        Key in `adata.obsm` to get the decipher v space.
    z_obsm_key: str
        Key in `adata.obsm` to get the decipher z space.
    v1_obs_col: str, optional
        Column name in `adata.obs` to align the first decipher component with.
        If None, only align the second component (or does not align at all if `v2` is also None).
    v1_order: list, optional
        List of values in `adata.obs[v1_col]`. The rotation will attempt to align these values in
        order along the v1 component. Must be provided if `adata.obs[v1_col]` is not numeric.
    v2_obs_col: str, optional
        Column name in `adata.obs` to align the second decipher component with.
        If None, only align the first component (or does not align at all if `v1` is also None).
    v2_order: list, optional
        List of values in `adata.obs[v2_col]`. The rotation will attempt to align these values in
        order along the v2 component. Must be provided if `adata.obs[v2_col]` is not numeric.
    auto_flip_decipher_z: bool, default True
        If True, flip each z to be correlated positively with the components.

    Returns
    -------
    `decipher_v`: ndarray
        The decipher v space after rotation.
    `decipher_z`: ndarray
        The decipher z space after flipping.
    `rotation`: ndarray
        The rotation matrix used to rotate the decipher v space.
    """

    def process_col_obs(v_col, v_order):
        if v_col is not None:
            v_obs = adata.obs[v_col]
            if v_order is not None:
                v_obs = v_obs.astype("category").cat.set_categories(v_order)
                v_obs = v_obs.cat.codes.replace(-1, np.nan)
            elif v_obs.dtype == "category":
                raise ValueError("v_order must be provided if v_obs is a category column")
            v_valid_cells = ~v_obs.isna()
            v_obs = v_obs[v_valid_cells].astype(float)
            return v_obs, v_valid_cells
        return None, None

    v1_obs, v1_valid_cells = process_col_obs(v1_obs_col, v1_order)
    v2_obs, v2_valid_cells = process_col_obs(v2_obs_col, v2_order)

    decipher_v = adata.obsm[v_obsm_key]

    def score_rotation(r):
        rotated_space = decipher_v @ r
        score = 0
        if v1_obs_col is not None:
            score += np.corrcoef(rotated_space[v1_valid_cells, 0], v1_obs)[1, 0]
            score -= np.abs(np.corrcoef(rotated_space[v1_valid_cells, 1], v1_obs)[1, 0])
        if v2_obs_col is not None:
            score += np.corrcoef(rotated_space[v2_valid_cells, 1], v2_obs)[1, 0]
            score -= np.abs(np.corrcoef(rotated_space[v2_valid_cells, 0], v2_obs)[1, 0])
        return score

    if v1_obs_col is None and v2_obs_col is None:
        raise ValueError("At least one of v1_obs_col or v2_obs_col must be provided")

    rotation_scores = []
    for t in np.linspace(0, 2 * np.pi, 100):
        for u in [1, -1]:
            rotation = _rot(t, u)
            rotation_scores.append((score_rotation(rotation), rotation))
    best_rotation = max(rotation_scores, key=lambda x: x[0])[1]

    decipher_v = decipher_v @ best_rotation

    decipher_z = adata.obsm[z_obsm_key]
    if auto_flip_decipher_z:
        # flip each z to be correlated positively with the components
        dim_z = decipher_z.shape[1]
        z_v_corr = np.corrcoef(decipher_z, y=decipher_v, rowvar=False)
        z_sign_correction = np.sign(z_v_corr[:dim_z, dim_z:].sum(axis=1))
        decipher_z = decipher_z * z_sign_correction

    return decipher_v, decipher_z, best_rotation



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import h5py
import numpy as np
import pandas as pd
from anndata.abc import CSCDataset, CSRDataset
from scipy.sparse import issparse
from torch.utils.data import Dataset

from scvi._constants import REGISTRY_KEYS

if TYPE_CHECKING:
    import torch

    from ._manager import AnnDataManager
from ._utils import registry_key_to_default_dtype, scipy_to_torch_sparse

logger = logging.getLogger(__name__)

SparseDataset = (CSRDataset, CSCDataset)


class AnnTorchDataset(Dataset):
    """Extension of :class:`~torch.utils.data.Dataset` for :class:`~anndata.AnnData` objects.

    Parameters
    ----------
    adata_manager
        :class:`~scvi.data.AnnDataManager` object with a registered AnnData object.
    getitem_tensors
        Specifies the keys in the data registry (``adata_manager.data_registry``) to return in
        ``__getitem__``. One of the following:

        * ``dict``: Keys correspond to keys in the data registry and values correspond to the
        desired :class:`~np.dtype` of the returned data.
        * ``list``: Elements correspond to keys in the data registry. Continuous data will be
        returned as :class:`~np.float32` and discrete data will be returned as :class:`~np.int64`.
        * ``None``: All registered data will be returned. Continuous data will be returned as
        :class:`~np.float32` and discrete data will be returned as :class:`~np.int64`.
    load_sparse_tensor
        ``EXPERIMENTAL`` If ``True``, loads data with sparse CSR or CSC layout as a
        :class:`~torch.Tensor` with the same layout. Can lead to speedups in data transfers to
        GPUs, depending on the sparsity of the data.
    """

    def __init__(
        self,
        adata_manager: AnnDataManager,
        getitem_tensors: list | dict[str, type] | None = None,
        load_sparse_tensor: bool = False,
    ):
        super().__init__()

        if adata_manager.adata is None:
            raise ValueError("Please run ``register_fields`` on ``adata_manager`` first.")
        self.adata_manager = adata_manager
        self.keys_and_dtypes = getitem_tensors
        self.load_sparse_tensor = load_sparse_tensor

    @property
    def registered_keys(self):
        """Keys in the data registry."""
        return self.adata_manager.data_registry.keys()

    @property
    def keys_and_dtypes(self):
        """Keys and corresponding :class:`~np.dtype` of data to fetch in ``__getitem__``."""
        return self._keys_and_dtypes

    @keys_and_dtypes.setter
    def keys_and_dtypes(self, getitem_tensors: list | dict[str, type] | None):
        """Set keys and corresponding :class:`~np.dtype` of data to fetch in ``__getitem__``.

        Raises an error if any of the keys are not in the data registry.
        """
        if isinstance(getitem_tensors, list):
            keys_to_dtypes = {key: registry_key_to_default_dtype(key) for key in getitem_tensors}
        elif isinstance(getitem_tensors, dict):
            keys_to_dtypes = getitem_tensors
        elif getitem_tensors is None:
            keys_to_dtypes = {
                key: registry_key_to_default_dtype(key) for key in self.registered_keys
            }
        else:
            raise ValueError("`getitem_tensors` must be a `list`, `dict`, or `None`")

        for key in keys_to_dtypes:
            if key not in self.registered_keys:
                raise KeyError(f"{key} not found in the data registry.")

        self._keys_and_dtypes = keys_to_dtypes

    @property
    def data(self):
        """Dictionary of data tensors.

        First time this is accessed, data is fetched from the underlying
        :class:`~anndata.AnnData` object. Subsequent accesses will return the
        cached dictionary.
        """
        if not hasattr(self, "_data"):
            self._data = {
                key: self.adata_manager.get_from_registry(key) for key in self.keys_and_dtypes
            }
        return self._data

    def __len__(self):
        return self.adata_manager.adata.shape[0]

    def __getitem__(
        self, indexes: int | list[int] | slice
    ) -> dict[str, np.ndarray | torch.Tensor]:
        """Fetch data from the :class:`~anndata.AnnData` object.

        Parameters
        ----------
        indexes
            Indexes of the observations to fetch. Can be a single index, a list of indexes, or a
            slice.

        Returns
        -------
        Mapping of data registry keys to arrays of shape ``(n_obs, ...)``.
        """
        if isinstance(indexes, int):
            indexes = [indexes]  # force batched single observations

        if self.adata_manager.adata.isbacked and isinstance(indexes, list | np.ndarray):
            # need to sort indexes for h5py datasets
            indexes = np.sort(indexes)

        data_map = {}

        for key, dtype in self.keys_and_dtypes.items():
            data = self.data[key]

            if isinstance(data, np.ndarray | h5py.Dataset):
                sliced_data = data[indexes].astype(dtype, copy=False)
            elif isinstance(data, pd.DataFrame):
                sliced_data = data.iloc[indexes, :].to_numpy().astype(dtype, copy=False)
            elif issparse(data) or isinstance(data, SparseDataset):
                sliced_data = data[indexes].astype(dtype, copy=False)
                if self.load_sparse_tensor:
                    sliced_data = scipy_to_torch_sparse(sliced_data)
                else:
                    sliced_data = sliced_data.toarray()
            elif isinstance(data, str) and key == REGISTRY_KEYS.MINIFY_TYPE_KEY:
                # for minified  anndata, we need this because we can have a string
                # for `data``, which is the value of the MINIFY_TYPE_KEY in adata.uns,
                # used to record the type data minification
                # TODO: Adata manager should have a list of which fields it will load
                continue
            else:
                raise TypeError(f"{key} is not a supported type")

            data_map[key] = sliced_data

        return data_map



from __future__ import annotations

from typing import TYPE_CHECKING

import flax.linen as nn
import jax
import jax.numpy as jnp
import numpyro.distributions as dist

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any, Literal

PYTORCH_DEFAULT_SCALE = 1 / 3


class Dense(nn.DenseGeneral):
    """Dense layer.

    Uses a custom initializer for the kernel to replicate the default PyTorch behavior.
    """

    def __init__(self, *args, **kwargs):
        from flax.linen.initializers import variance_scaling

        _kwargs = {"kernel_init": variance_scaling(PYTORCH_DEFAULT_SCALE, "fan_in", "uniform")}
        _kwargs.update(kwargs)

        super().__init__(*args, **_kwargs)


class ResnetBlock(nn.Module):
    """Resnet block.

    Consists of the following operations:

    1. :class:`~flax.linen.Dense`
    2. :class:`~flax.linen.LayerNorm`
    3. Activation function specified by ``internal_activation``
    4. Skip connection if ``n_in`` is equal to ``n_hidden``, otherwise a :class:`~flax.linen.Dense`
        layer is applied to the input before the skip connection to match features.
    5. :class:`~flax.linen.Dense`
    6. :class:`~flax.linen.LayerNorm`
    7. Activation function specified by ``output_activation``

    Parameters
    ----------
    n_out
        Number of output units.
    n_hidden
        Number of hidden units.
    internal_activation
        Activation function to use after the first :class:`~flax.linen.Dense` layer.
    output_activation
        Activation function to use after the last :class:`~flax.linen.Dense` layer.
    training
        Whether the model is in training mode.
    """

    n_out: int
    n_hidden: int = 128
    internal_activation: Callable[[jax.typing.ArrayLike], jax.Array] = nn.relu
    output_activation: Callable[[jax.typing.ArrayLike], jax.Array] = nn.relu
    training: bool | None = None

    @nn.compact
    def __call__(self, inputs: jax.typing.ArrayLike, training: bool | None = None) -> jax.Array:
        training = nn.merge_param("training", self.training, training)
        h = Dense(self.n_hidden)(inputs)
        h = nn.LayerNorm()(h)
        h = self.internal_activation(h)
        if inputs.shape[-1] != self.n_hidden:
            h = h + Dense(self.n_hidden)(inputs)
        else:
            h = h + inputs
        h = Dense(self.n_out)(h)
        h = nn.LayerNorm()(h)
        return self.output_activation(h)


class MLP(nn.Module):
    """Multi-layer perceptron with resnet blocks.

    Applies ``n_layers`` :class:`~ResnetBlock` blocks to the input, followed by a
    :class:`~flax.linen.Dense` layer to project to the output dimension.

    Parameters
    ----------
    n_out
        Number of output units.
    n_hidden
        Number of hidden units.
    n_layers
        Number of resnet blocks.
    activation
        Activation function to use.
    training
        Whether the model is in training mode.
    """

    n_out: int
    n_hidden: int = 128
    n_layers: int = 1
    activation: Callable[[jax.typing.ArrayLike], jax.Array] = nn.relu
    training: bool | None = None

    @nn.compact
    def __call__(self, inputs: jax.typing.ArrayLike, training: bool | None = None) -> jax.Array:
        training = nn.merge_param("training", self.training, training)
        h = inputs
        for _ in range(self.n_layers):
            h = ResnetBlock(
                n_out=self.n_hidden,
                internal_activation=self.activation,
                output_activation=self.activation,
            )(h, training=training)
        return Dense(self.n_out)(h)


class NormalDistOutputNN(nn.Module):
    """Fully-connected neural net parameterizing a normal distribution.

    Applies ``n_layers`` :class:`~ResnetBlock` blocks to the input, followed by a
    :class:`~flax.linen.Dense` layer for the mean and a :class:`~flax.linen.Dense` and
    :func:`~flax.linen.softplus` layer for the scale.

    Parameters
    ----------
    n_out
        Number of output units.
    n_hidden
        Number of hidden units.
    n_layers
        Number of resnet blocks.
    scale_eps
        Numerical stability constant added to the scale of the normal distribution.
    """

    n_out: int
    n_hidden: int = 128
    n_layers: int = 1
    scale_eps: float = 1e-5
    training: bool | None = None

    @nn.compact
    def __call__(self, inputs: jax.typing.ArrayLike, training: bool | None = None) -> dist.Normal:
        training = nn.merge_param("training", self.training, training)
        h = inputs
        for _ in range(self.n_layers):
            h = ResnetBlock(n_out=self.n_hidden)(h, training=training)
        mean = Dense(self.n_out)(h)
        scale = nn.Sequential([Dense(self.n_out), nn.softplus])(h)
        return dist.Normal(mean, scale + self.scale_eps)


class ConditionalNormalization(nn.Module):
    """Condition-specific normalization.

    Applies either batch normalization or layer normalization to the input, followed by
    condition-specific scaling (``gamma``) and shifting (``beta``).

    Parameters
    ----------
    n_features
        Number of features.
    n_conditions
        Number of conditions.
    training
        Whether the model is in training mode.
    normalization_type
        Type of normalization to apply. Must be one of ``"batch", "layer"``.
    """

    n_features: int
    n_conditions: int
    training: bool | None = None
    normalization_type: Literal["batch", "layer"] = "layer"

    @staticmethod
    def _gamma_initializer() -> jax.nn.initializers.Initializer:
        def init(key: jax.random.KeyArray, shape: tuple, dtype: Any = jnp.float_) -> jax.Array:
            weights = jax.random.normal(key, shape, dtype) * 0.02 + 1
            return weights

        return init

    @staticmethod
    def _beta_initializer() -> jax.nn.initializers.Initializer:
        def init(key: jax.random.KeyArray, shape: tuple, dtype: Any = jnp.float_) -> jax.Array:
            del key
            weights = jnp.zeros(shape, dtype=dtype)
            return weights

        return init

    @nn.compact
    def __call__(
        self,
        x: jax.typing.ArrayLike,
        condition: jax.typing.ArrayLike,
        training: bool | None = None,
    ) -> jax.Array:
        training = nn.merge_param("training", self.training, training)

        if self.normalization_type == "batch":
            x = nn.BatchNorm(use_bias=False, use_scale=False)(x, use_running_average=not training)
        elif self.normalization_type == "layer":
            x = nn.LayerNorm(use_bias=False, use_scale=False)(x)
        else:
            raise ValueError("`normalization_type` must be one of ['batch', 'layer'].")

        cond_int = condition.squeeze(-1).astype(int)
        gamma = nn.Embed(
            self.n_conditions,
            self.n_features,
            embedding_init=self._gamma_initializer(),
            name="gamma_conditional",
        )(cond_int)
        beta = nn.Embed(
            self.n_conditions,
            self.n_features,
            embedding_init=self._beta_initializer(),
            name="beta_conditional",
        )(cond_int)

        return gamma * x + beta


class AttentionBlock(nn.Module):
    """Attention block consisting of multi-head self-attention and MLP.

    Parameters
    ----------
    query_dim
        Dimension of the query input.
    out_dim
        Dimension of the output.
    outerprod_dim
        Dimension of the outer product.
    n_channels
        Number of channels.
    n_heads
        Number of heads.
    dropout_rate
        Dropout rate.
    n_hidden_mlp
        Number of hidden units in the MLP.
    n_layers_mlp
        Number of layers in the MLP.
    training
        Whether the model is in training mode.
    stop_gradients_mlp
        Whether to stop gradients through the MLP.
    activation
        Activation function to use.
    """

    query_dim: int
    out_dim: int
    outerprod_dim: int = 16
    n_channels: int = 4
    n_heads: int = 2
    dropout_rate: float = 0.0
    n_hidden_mlp: int = 32
    n_layers_mlp: int = 1
    training: bool | None = None
    stop_gradients_mlp: bool = False
    activation: Callable[[jax.Array], jax.Array] = nn.gelu

    @nn.compact
    def __call__(
        self,
        query_embed: jax.typing.ArrayLike,
        kv_embed: jax.typing.ArrayLike,
        training: bool | None = None,
    ) -> jax.Array:
        training = nn.merge_param("training", self.training, training)
        has_mc_samples = query_embed.ndim == 3

        query_embed_stop = (
            query_embed if not self.stop_gradients_mlp else jax.lax.stop_gradient(query_embed)
        )
        query_for_att = nn.DenseGeneral((self.outerprod_dim, 1), use_bias=False)(query_embed_stop)
        kv_for_att = nn.DenseGeneral((self.outerprod_dim, 1), use_bias=False)(kv_embed)
        eps = nn.MultiHeadDotProductAttention(
            num_heads=self.n_heads,
            qkv_features=self.n_channels * self.n_heads,
            out_features=self.n_channels,
            dropout_rate=self.dropout_rate,
            use_bias=True,
        )(inputs_q=query_for_att, inputs_kv=kv_for_att, deterministic=not training)

        if not has_mc_samples:
            eps = jnp.reshape(eps, (eps.shape[0], eps.shape[1] * eps.shape[2]))
        else:
            eps = jnp.reshape(eps, (eps.shape[0], eps.shape[1], eps.shape[2] * eps.shape[3]))

        eps_ = MLP(
            n_out=self.outerprod_dim,
            n_hidden=self.n_hidden_mlp,
            training=training,
            activation=self.activation,
        )(inputs=eps)
        inputs = jnp.concatenate([query_embed, eps_], axis=-1)
        residual = MLP(
            n_out=self.out_dim,
            n_hidden=self.n_hidden_mlp,
            n_layers=self.n_layers_mlp,
            training=training,
            activation=self.activation,
        )(inputs=inputs)
        return residual



import logging

import numpy as np

from scvi.data import AnnDataManager
from scvi.data._constants import _SETUP_ARGS_KEY
from scvi.external.methylvi import METHYLVI_REGISTRY_KEYS

logger = logging.getLogger(__name__)


def scmc_raw_counts_properties(
    adata_manager: AnnDataManager,
    idx1: list[int] | np.ndarray,
    idx2: list[int] | np.ndarray,
    var_idx: list[int] | np.ndarray | None = None,
    context: str = None,
) -> dict[str, np.ndarray]:
    """Computes and returns some statistics on the raw counts of two sub-populations.

    Parameters
    ----------
    adata_manager
        :class:`~scvi.data.AnnDataManager` object setup with :class:`~scvi.model.SCVI`.
    idx1
        subset of indices describing the first population.
    idx2
        subset of indices describing the second population.
    mc_key
        denotes layer where methylated cytosine counts are stored in AnnData
    cov_key
        denotes layer where total cytosine counts are stored in AnnData
    var_idx
        subset of variables to extract properties from. if None, all variables are used.

    Returns
    -------
    type
        Dict of ``np.ndarray`` containing, by pair (one for each sub-population).
    """
    adata = adata_manager.adata
    mc = adata[context].layers[adata_manager.registry[_SETUP_ARGS_KEY]["mc_layer"]]
    cov = adata[context].layers[adata_manager.registry[_SETUP_ARGS_KEY]["cov_layer"]]
    mc1 = mc[idx1]
    mc2 = mc[idx2]

    cov1 = cov[idx1]
    cov2 = cov[idx2]
    if var_idx is not None:
        mc1 = mc1[:, var_idx]
        mc2 = mc2[:, var_idx]

        cov1 = cov1[:, var_idx]
        cov2 = cov2[:, var_idx]

    mean1 = np.asarray(np.nanmean(mc1 / cov1, axis=0)).ravel()
    mean2 = np.asarray(np.nanmean(mc2 / cov2, axis=0)).ravel()

    properties = {"emp_mean1": mean1, "emp_mean2": mean2, "emp_effect": (mean1 - mean2)}
    return properties


def _context_cov_key(context):
    return f"{context}_{METHYLVI_REGISTRY_KEYS.COV_KEY}"


def _context_mc_key(context):
    return f"{context}_{METHYLVI_REGISTRY_KEYS.MC_KEY}"
